"""Display signals into a grid..

A large portion of this code was taken from the example developped by the vispy
team :
https://github.com/vispy/vispy/blob/master/examples/demo/gloo/realtime_signals.py
"""
import numpy as np
from itertools import product

from vispy import gloo, visuals
from vispy.scene.visuals import create_visual_node, Text

from sleep_scoring.utils import color2vb, vispy_array, PrepareData, ndsubplot


__all__ = ('GridSignal')


vertex_shader = """
#version 120
varying vec3 v_index;
varying vec4 v_color;
// Varying variables used for clipping in the fragment shader.
varying vec2 v_position;
varying vec4 v_ab;
varying vec3 a_pos;
void main() {
    float nrows = $u_size.x;
    float ncols = $u_size.y;

    // Compute the x coordinate from the time index.
    float x = -1 + 2*$a_index.z / ($u_n-1);

    // Turn position into a vec3 :
    a_pos = vec3($a_position, 1, 1);
    vec2 position = vec2(x - (1 - 1 / $u_scale.x), a_pos);

    // Find the affine transformation for the subplots.
    vec2 a = vec2(1./ncols, 1./nrows)*.98;
    vec2 b = vec2(-1 + $u_space*($a_index.x+.5) / ncols,
                  -1 + $u_space*($a_index.y+.5) / nrows);

    // Apply the static subplot transformation + scaling.
    gl_Position = $transform(vec4(a*$u_scale*position+b, 0.0, 1.0));

    // Use texture 1D to get the color.
    v_color = texture1D($u_text, $a_color);

    // For clipping test in the fragment shader.
    v_index = $a_index;
    v_position = gl_Position.xy;
}
"""

fragment_shader = """
#version 120
varying vec4 v_color;
varying vec3 v_index;
varying vec2 v_position;
void main() {
    gl_FragColor = vec4(v_color.xyz, 1.);

    // Discard the fragments between the signals (emulate glMultiDrawArrays).
    if ((fract(v_index.x) > 0.) || (fract(v_index.y) > 0.))
        discard;
}
"""


class GridSignalVisual(visuals.Visual):
    """Visual class for grid of signals.

    Parameters
    ----------
    data : array_like
        Array of data. Could be 1-D, 2-D or 3-D.
    axis : int | -1
        Time axis location.
    sf : float | 1.
        The sampling frequency (used for filtering).
    random : array_like/string/tuple | 'random'
        Use 'random' for random colors or a color name for uniform color.
    space : float | 2.
        Space between subplots.
    scale : tuple | (1., 1.)
        Tuple descigin the scaling along the x and y-axis.
    """

    def __len__(self):
        """Return the number of time points."""
        return self._n

    def __init__(self, data, axis=-1, sf=1., color='random', title=None,
                 space=2., scale=(.98, .9), font_size=10., width=1.,
                 method='gl', force_shape=None):
        """Init."""
        # =========================== CHECKING ===========================
        assert isinstance(data, np.ndarray) and (data.ndim <= 3)
        assert isinstance(axis, int)
        assert isinstance(sf, (int, float))
        assert isinstance(space, (int, float))
        assert isinstance(scale, (tuple, list)) and len(scale) == 2

        # =========================== VISUALS ===========================
        visuals.Visual.__init__(self, vertex_shader, fragment_shader)

        self.set_gl_state('translucent', depth_test=True, cull_face=False,
                          blend=True, blend_func=('src_alpha',
                                                  'one_minus_src_alpha'))
        self._draw_mode = 'line_strip'
        self._txt = Text(bold=False, font_size=font_size)

        # =========================== DATA ===========================
        # Keep some inputs :
        self._sh = data.shape
        self._n = self._sh[axis]
        self._axis = axis
        self._sf = sf
        self._color = color
        self.scale = scale
        self.space = space
        self._prep = PrepareData(axis=-1)
        self.width = width
        self.method = method

        # =========================== BUFFERS ===========================
        # Create buffers (for data, index and color)
        rnd_1 = np.zeros((3,), dtype=np.float32)
        rnd_3 = np.zeros((1, 3), dtype=np.float32)
        self._dbuffer = gloo.VertexBuffer(rnd_1)
        self._ibuffer = gloo.VertexBuffer(rnd_3)
        self._cbuffer = gloo.VertexBuffer()
        # Send to the program :
        self.shared_program.vert['a_position'] = self._dbuffer
        self.shared_program.vert['a_index'] = self._ibuffer
        self.shared_program.vert['u_size'] = (1, 1)
        self.shared_program.vert['u_n'] = len(self)

        # Set data :
        self.set_data(data, axis, color, title, force_shape)
        self.freeze()

    def set_data(self, data=None, axis=None, color=None, title=None,
                 force_shape=None):
        """Set data to the grid of signals.

        Parameters
        ----------
        data : None
            Array of data. Could be 1-D, 2-D or 3-D.
        axis : int | None
            Time axis location.
        random : array_like/string/tuple | 'random'
            Use 'random' for random colors or a color name for uniform color.
        """
        rnd_dyn = (.2, .8)  # random color range
        # ====================== CHECKING ======================
        # Axis :
        axis = axis if isinstance(axis, int) else self._axis
        axis = len(self._sh) - 1 if axis == -1 else axis

        # ====================== CHECKING ======================
        # Data :
        if isinstance(data, np.ndarray):
            # -------------- (n_rows, n_cols, n_time) --------------
            if data.ndim == 1:  # 1-D array
                data = data.reshape(1, 1, -1)
                g_size = (1, 1)
            elif data.ndim == 2:  # 2-D array
                if axis == 0:  # data need to be transposed
                    data = np.swapaxes(data, 0, 1)
                    axis = 1
                g_size = (data.shape[0], 1)  # (n_row, 1)
                data = data[np.newaxis, ...]
            elif data.ndim == 3:  # 3-D array
                if axis != data.ndim - 1:  # data need to be transposed
                    data = np.swapaxes(data, axis, -1)
                    axis = data.ndim - 1
                g_size = (data.shape[0], data.shape[1])

            # -------------- Signals index --------------
            m = np.prod(list(data.shape)[0:-1])
            sig_index = np.arange(m).reshape(*g_size)

            # -------------- Optimal 2-D --------------
            self._data = data
            self._ori_shape = list(data.shape)[0:-1]
            if force_shape is None:
                n_rows, n_cols = ndsubplot(m)
            elif len(g_size) == 2:
                n_rows, n_cols = force_shape
            data = data.reshape(n_rows, n_cols, len(self))
            sig_index = sig_index.reshape(n_rows, n_cols)
            g_size = (n_rows, n_cols)
            self._opt_shape = list(data.shape)[0:-1]
            self._sig_index = sig_index

            # -------------- (n_rows * n_cols, n_time) --------------
            data = np.reshape(data, (m, len(self)), order='F')

            # -------------- Prepare --------------
            # Force demean / detrend of _prep :
            self._prep.demean, self._prep.detrend = False, False
            data = self._prep._prepare_data(self._sf, data, 0)
            # Demean and normalize :
            kw = {'axis': -1, 'keepdims': True}
            data -= data.mean(**kw)
            data /= np.abs(data).max(**kw)
            self._dbuffer.set_data(vispy_array(data))
            self.g_size = g_size

        # ====================== INDEX ======================
        n, m = len(self), np.prod(g_size)
        self._sig_index = self._sig_index.reshape(n_rows, n_cols)
        idg = np.c_[np.repeat(np.repeat(np.arange(n_cols), n_rows), n),
                    np.repeat(np.tile(np.arange(n_rows), n_cols), n)[::-1],
                    np.tile(np.arange(n), m)].astype(np.float32)
        self._ibuffer.set_data(vispy_array(idg))

        # ====================== COLOR ======================
        if color is not None:
            g_size = np.array(self.g_size)
            n = len(self)
            if color == 'random':  # (m, 3) random color
                color_1d = np.random.uniform(size=(m, 3), low=rnd_dyn[0],
                                             high=rnd_dyn[1])
                color_idx = np.mgrid[0:m, 0:len(self)][0] / m
            elif color is not None:  # (m, 3) uniform color
                color_1d = color2vb(color)
                color_idx = np.zeros((m * len(self)), dtype=np.float32)
            # Send texture to vertex shader :
            text_1d = gloo.Texture1D(color_1d.astype(np.float32))
            self.shared_program.vert['u_text'] = text_1d
            # Send color index to use for the texture :
            self._cbuffer.set_data(color_idx.astype(np.float32))
            self.shared_program.vert['a_color'] = self._cbuffer

        # ====================== TITLES ======================
        # Titles checking :
        if title is None or (len(title) != m):
            st, it = '({}, {})', product(range(n_rows), range(n_cols))
            title = [st.format(i, k) for i, k in it]
        # Set text and font size :
        if not self._txt.text:
            self._txt.text = title
        # Get titles position :
        x_factor, y_factor = 1. / (n_cols), 1. / (n_rows)
        r_x = np.linspace(-1. + x_factor, 1. - x_factor, n_cols)
        r_x = np.tile(r_x, n_rows)
        r_y = np.linspace(-1. + y_factor, 1. - y_factor, n_rows)[::-1]
        r_y += y_factor
        r_y = np.repeat(r_y, n_cols)
        pos = np.c_[r_x, r_y, np.full_like(r_x, -10.)]
        self._txt.pos = pos.astype(np.float32)

    def clean(self):
        """Clean buffers."""
        self._dbuffer.delete()
        self._ibuffer.delete()
        self._cbuffer.delete()

    def _convert_row_cols(self, row, col):
        """Convert row and col according to the optimal grid."""
        try:
            index = self._sig_index[row, col]
            idx = np.where(self._sig_index.reshape(*self._ori_shape) == index)
            return idx[0][0], idx[1][0]
        except:
            return row, col

    def _prepare_transforms(self, view):
        """Call for the first rendering."""
        tr = view.transforms
        view_vert = view.view_program.vert
        view_vert['transform'] = tr.get_transform()

    def _prepare_draw(self, view=None):
        """Function called everytime there's a camera update."""
        try:
            import OpenGL.GL as GL
            GL.glLineWidth(self._width)
            if self._smooth_line:
                GL.glEnable(GL.GL_LINE_SMOOTH)
            else:
                GL.glDisable(GL.GL_LINE_SMOOTH)
        except Exception:  # can be other than ImportError sometimes
            pass

    # ========================================================================
    # ========================================================================
    # PROPERTIES
    # ========================================================================
    # ========================================================================
    # ----------- SCALE -----------
    @property
    def scale(self):
        """Get the scale value."""
        return self._scale

    @scale.setter
    def scale(self, value):
        """Set scale value."""
        if isinstance(value, tuple):
            self._uscale = value
            self._scale = value
            self.shared_program.vert['u_scale'] = value
            self.update()

    # ----------- SPACE -----------
    @property
    def space(self):
        """Get the space value."""
        return self._space

    @space.setter
    def space(self, value):
        """Set space value."""
        if isinstance(value, float):
            self._space = value
            self.shared_program.vert['u_space'] = value
            self.update()

    # ----------- RECT -----------
    @property
    def rect(self):
        return (-1.05, -1.1, self._space + .1, self._space + .2)

    # ----------- FONT_SIZE -----------
    @property
    def font_size(self):
        """Get the font_size value."""
        return self._txt.font_size

    @font_size.setter
    def font_size(self, value):
        """Set font_size value."""
        self._txt.font_size = value

    # ----------- TCOLOR -----------
    @property
    def tcolor(self):
        """Get the tcolor value."""
        return self._txt.color

    @tcolor.setter
    def tcolor(self, value):
        """Set tcolor value."""
        self._txt.color = color2vb(value)

    # ----------- TVISIBLE -----------
    @property
    def tvisible(self):
        """Get the tvisible value."""
        return self._txt.visible

    @tvisible.setter
    def tvisible(self, value):
        """Set tvisible value."""
        self._txt.visible = value

    # ----------- G_SIZE -----------
    @property
    def g_size(self):
        """Get the g_size value."""
        return self._g_size

    @g_size.setter
    def g_size(self, value):
        """Set g_size value."""
        self._g_size = value
        self.shared_program.vert['u_size'] = value
        self.update()

    # ----------- WIDTH -----------
    @property
    def width(self):
        """Get the width value."""
        return self._width

    @width.setter
    def width(self, value):
        """Set width value."""
        self._width = value
        self.update()

    # ----------- METHOD -----------
    @property
    def method(self):
        """Get the method value."""
        return self._method

    @method.setter
    def method(self, value):
        """Set method value."""
        self._method = value
        self._smooth_line = value == 'agg'
        self.update()


GridSignal = create_visual_node(GridSignalVisual)



"""Main class for Visbrain objects."""
import os
import sys
import logging

import vispy
import vispy.visuals.transforms as vist

from .scene_obj import VisbrainCanvas
from ..io import (write_fig_canvas, dialog_save, path_to_visbrain_data,
                  load_config_json, get_data_url_path, download_file,
                  get_files_in_folders)
from ..utils import color2vb, set_log_level, merge_cameras
from ..config import CONFIG
from ..visuals import CbarBase

logger = logging.getLogger('visbrain')


class _VisbrainShortcuts(object):
    """Add shortcuts to a canvas."""

    def __init__(self):
        """Init."""
        self.key_press = {}
        self.events = ['mouse_release', 'mouse_double_click', 'mouse_move',
                       'mouse_press']

    def set_shortcuts_to_canvas(self, canvas):
        """Set shortcuts to a VisbrainCanvas."""
        assert isinstance(canvas, VisbrainCanvas)
        self.canvas = canvas
        ce = canvas.canvas.events  # noqa
        for k in self.events:
            if hasattr(self, '_on_' + k):
                eval('ce.%s.connect(self._on_%s())' % (k, k))
        # Key-pressed :

        def _save_canvas(event):
            from PyQt5.QtWidgets import QWidget
            ext = ['png', 'tiff', 'jpg']
            _ext = ['%s file (*.%s)' % (k.upper(), k) for k in ext]
            _ext += ['All files (*.*)']
            saveas = dialog_save(QWidget(), name='Export the scene',
                                 default='canvas.png', allext=_ext)
            if saveas:
                write_fig_canvas(saveas, self.canvas.canvas,
                                 widget=self.canvas.canvas.central_widget)
        self.key_press['s'] = _save_canvas

        def _fcn_key_press(event):
            """Key pressed."""
            if event.text in self.key_press.keys():
                self.key_press[event.text](event)
        ce.key_press.connect(_fcn_key_press)


class _VisbrainObj(CbarBase, _VisbrainShortcuts):
    """Class for VisbrainObjects and CombineObjects."""

    def __init__(self, **kw):
        """Init."""
        CbarBase.__init__(self, **kw)
        _VisbrainShortcuts.__init__(self)
        self._default_cblabel = ''
        self._data_folder = None

    # --------------------------- CAMERA ---------------------------
    def _get_camera(self):
        raise NotImplementedError

    # --------------------------- COLORBAR ---------------------------
    def _update_cbar(self):
        raise NotImplementedError

    def _update_cbar_minmax(self):
        raise NotImplementedError

    # --------------------------- DATA ---------------------------
    def _df_is_downloadable(self, file):
        """Get if a file name could be downloaded."""
        json_path = get_data_url_path()
        json_struct = load_config_json(json_path)[self._data_folder]
        return file in json_struct

    def _df_is_downloaded(self, file):
        """Get if a file has already been downloaded."""
        return os.path.isfile(os.path.join(self._data_folder_path, file))

    def _df_get_downloadable(self):
        """Get the list of files that can be downloaded."""
        json_path = get_data_url_path()
        return load_config_json(json_path)[self._data_folder]

    def _df_get_tmp_folder(self):
        """Get the tmp associated folder."""
        vb_path = path_to_visbrain_data()
        return os.path.join(*(vb_path, 'tmp', self._data_folder))

    def _df_get_downloaded(self, **kwargs):
        """Get the list of files that are already downloaded."""
        main_path = get_files_in_folders(self._data_folder_path, **kwargs)
        tmp_path = self._df_get_tmp_folder()
        if os.path.isdir(tmp_path):
            main_path += get_files_in_folders(tmp_path, **kwargs)
        return main_path

    def _df_get_file(self, file, download=True):
        """Get the path to a file or download it if needed."""
        is_dl = self._df_is_downloaded(file)
        if not is_dl and download:
            assert self._df_is_downloadable(file)
            self._df_download_file(file)
        # Find if the file is in _data_folder or in tmp :
        if file in os.listdir(self._data_folder_path):
            use_path = self._data_folder_path
        else:
            use_path = self._df_get_tmp_folder()
        return os.path.join(use_path, file)

    def _df_download_file(self, file):
        """Download a file."""
        return download_file(file, astype=self._data_folder)

    # ----------- DATA_FOLDER -----------
    @property
    def data_folder(self):
        """Get the data_folder value."""
        return self._data_folder

    @data_folder.setter
    def data_folder(self, value):
        """Set data_folder value."""
        if isinstance(self._data_folder, str):
            raise ValueError("data_folder can only be set once.")
        assert isinstance(value, str)
        # Create the directory if it doesn't exist :
        full_path = path_to_visbrain_data(folder=value)
        if not os.path.exists(full_path):
            os.makedirs(full_path)
            logger.info("%s folder created" % full_path)
        self._data_folder = value
        self._data_folder_path = full_path


class VisbrainObject(_VisbrainObj):
    """Base class inherited by all of the Visbrain objects.

    Parameters
    ----------
    name : string
        Object name.
    parent : VisPy.parent | None
        Markers object parent.
    transform : VisPy.visuals.transforms | None
        VisPy transformation to set to the parent node.
    verbose : string
        Verbosity level.
    """

    def __init__(self, name, parent=None, transform=None, verbose=None, **kw):
        """Init."""
        _VisbrainObj.__init__(self, **kw)
        self._node = vispy.scene.Node(name=name)
        self._node.parent = parent
        self._csize = None  # canvas size
        self._shortcuts = {}
        # Name :
        assert isinstance(name, str)
        self._name = name
        # Transformation :
        if transform is None:
            transform = vist.STTransform()
        self._node.transform = transform
        # Verbose :
        set_log_level(verbose)
        logger.info('%s created' % repr(self))

    def __repr__(self):
        """Represent ClassName(name='object_name')."""
        return type(self).__name__ + "(name='" + self._name + "')"

    def __str__(self):
        """Return the object name."""
        return self._name

    def _get_parent(self, bgcolor, axis, show, obj=None, **kwargs):
        """Get the object parent for preview and screenshot."""
        if hasattr(obj, '_get_camera'):
            camera = obj._get_camera()
        else:
            camera = self._get_camera()
        canvas = VisbrainCanvas(axis=axis, show=show, name=self._name,
                                bgcolor=color2vb(bgcolor), camera=camera,
                                shortcuts=self._shortcuts, **kwargs)
        self._csize = canvas.canvas.size
        self.set_shortcuts_to_canvas(canvas)
        if not hasattr(self._node.parent, 'name'):
            self._node.parent = canvas.wc.scene
        return canvas

    def preview(self, bgcolor='black', axis=False, xyz=False, show=True,
                obj=None, size=(1200, 800), **kwargs):
        """Previsualize the result.

        Parameters
        ----------
        bgcolor : array_like/string/tuple | 'black'
            Background color for the preview.
        axis : bool | False
            Add x and y axis with ticks.
        xyz : bool | False
            Add an (x, y, z) axis to the scene.
        obj : VisbrainObj | None
            Pass a Visbrain object if you want to use the camera of an other
            object.
        size : tuple | (1200, 800)
            Default size of the window.
        kwargs : dict | {}
            Optional arguments are passed to the VisbrainCanvas class.
        """
        parent_bck = self._node.parent
        kwargs['cargs'] = {'size': size}
        canvas = self._get_parent(bgcolor, axis, show, obj, **kwargs)
        if xyz:
            vispy.scene.visuals.XYZAxis(parent=canvas.wc.scene)
        # view.camera = camera
        if (sys.flags.interactive != 1) and show:
            CONFIG['VISPY_APP'].run()
        # Reset orignial parent :
        self._node.parent = parent_bck

    def describe_tree(self):
        """Tree description."""
        return self._node.describe_tree()

    def screenshot(self, saveas, print_size=None, dpi=300., unit='centimeter',
                   factor=None, region=None, autocrop=False, bgcolor=None,
                   transparent=False, obj=None, line_width=1., **kwargs):
        """Take a screeshot of the scene.

        By default, the rendered canvas will have the size of your screen.
        The screenshot() method provides two ways to increase to exported image
        resolution :

            * Using print_size, unit and dpi inputs : specify the size of the
              image at a specific dpi level. For example, you might want to
              have an (10cm, 15cm) image at 300 dpi.
            * Using the factor input : multiply the default image size by this
              factor. For example, if you have a (1920, 1080) monitor and if
              factor is 2, the exported image should have a shape of
              (3840, 2160) pixels.

        Parameters
        ----------
        saveas : str
            The name of the file to be saved. This file must contains a
            extension like .png, .tiff, .jpg...
        print_size : tuple | None
            The desired print size. This argument should be used in association
            with the dpi and unit inputs. print_size describe should be a tuple
            of two floats describing (width, height) of the exported image for
            a specific dpi level. The final image might not have the exact
            desired size but will try instead to find a compromize
            regarding to the proportion of width/height of the original image.
        dpi : float | 300.
            Dots per inch for printing the image.
        unit : {'centimeter', 'millimeter', 'pixel', 'inch'}
            Unit of the printed size.
        factor : float | None
            If you don't want to use the print_size input, factor simply
            multiply the resolution of your screen.
        region : tuple | None
            Select a specific region. Must be a tuple of four integers each one
            describing (x_start, y_start, width, height).
        autocrop : bool | False
            Automaticaly crop the figure in order to have the smallest
            space between the brain and the border of the picture.
        bgcolor : array_like/string | None
            The background color of the image.
        transparent : bool | False
            Specify if the exported figure have to contains a transparent
            background.
        obj : VisbrainObj | None
            Pass a Visbrain object if you want to use the camera of an other
            object for the sceen rendering.
        kwargs : dict | {}
            Optional arguments are passed to the VisbrainCanvas class.
        """
        kw = dict(print_size=print_size, dpi=dpi, factor=factor,
                  autocrop=autocrop, unit=unit, region=region,
                  bgcolor=bgcolor, transparent=transparent)
        canvas = self._get_parent(bgcolor, False, False, obj, **kwargs)
        write_fig_canvas(saveas, canvas.canvas,
                         widget=canvas.canvas.central_widget, **kw)
        self._node.parent = None

    def copy(self):
        """Get a copy of the object."""
        from copy import copy
        return copy(self)

    # ----------- PARENT -----------
    @property
    def parent(self):
        """Get the parent value."""
        p_isnn = self._node.parent is not None
        return self._node.parent if p_isnn else self._node

    @parent.setter
    def parent(self, value):
        """Set parent value."""
        self._node.parent = value

    # ----------- TRANSFORM -----------
    @property
    def transform(self):
        """Get the transform value."""
        return self._node.transform

    @transform.setter
    def transform(self, value):
        """Set transform value."""
        self._node.transform = value

    # ----------- NAME -----------
    @property
    def name(self):
        """Get the name value."""
        return self._name

    # ----------- VISIBLE_OBJ -----------
    @property
    def visible_obj(self):
        """Get the visible_obj value."""
        return self._node.visible

    @visible_obj.setter
    def visible_obj(self, value):
        """Set visible_obj value."""
        assert isinstance(value, bool)
        self._node.visible = value


class CombineObjects(_VisbrainObj):
    """Combine Visbrain objects.

    Parameters
    ----------
    obj_type : Object type
        The object type.
    objects : list or obj_type
        List of objects (or single object).
    """

    def __init__(self, obj_type, objects, select=None, parent=None):
        """Init."""
        _VisbrainObj.__init__(self)
        # Parent node for combined objects :
        self._cnode = vispy.scene.Node(name='Combine ' + obj_type.__name__)
        self._cnode.parent = parent
        # Initialize objects :
        self._objs, self._objs_order = {}, []
        self._visible_obj = True
        self._obj_type = obj_type.__name__
        if objects is not None:
            if isinstance(objects, obj_type):  # single object
                objects = [objects]
            if isinstance(objects, list):  # list of objects
                assert all([isinstance(k, obj_type) for k in objects])
                self._objs_order = [k.name for k in objects]
                self._objs = {k.name: k for k in objects}
            else:
                raise TypeError("Wrong object type. Can not combine objects.")
            # Select an object. By default, the first one is selected :
            select = self._objs_order[0] if select is None else select
            self.select(select)
            self.name = str(self)
            # Set parent :
            for k in self:
                k.parent = self._cnode
        else:
            self.name = None

    def __getitem__(self, value):
        """Get the object from the key value."""
        if isinstance(value, str):
            return self._objs[value]
        elif isinstance(value, int):
            return self._objs[self._objs_order[value]]

    def __setitem__(self, key, value):
        """Set an attribute of the selected object.."""
        exec('self[self._select].' + key + ' = value')

    def __len__(self):
        """Get the number of objects."""
        return len(self._objs_order)

    def __iter__(self):
        """Iterate over objects."""
        for k in self._objs_order:
            yield self[k]

    def __str__(self):
        """Get the name of all objects."""
        return ' + '.join(self._objs_order)

    def __repr__(self):
        """Represent combined objects."""
        reprs = [repr(k) for k in self]
        return type(self).__name__ + "(" + ", ".join(reprs) + ")"

    def _get_camera(self):
        """Return a merged version of cameras."""
        cams = []
        for k in self:
            cams.append(k._get_camera())
        return merge_cameras(*tuple(cams))

    def update(self):
        """Update every objects."""
        self._cnode.update()
        for k in self:
            k._node.update()
            k.update()

    def get_list_of_objects(self):
        """Get the list of defined objects."""
        return self._objs_order

    def get_selected_object(self):
        """Get the curently selected object."""
        return self._select

    def append(self, obj):
        """Add a new object."""
        assert type(obj).__name__ == self._obj_type
        self._objs_order.append(obj.name)
        self._objs[obj.name] = obj

    def select(self, name=None):
        """Select an object.

        Parameters
        ----------
        name : string | None
            Name of the object to select.
        """
        if name is not None:
            assert name in self._objs_order
            self._select = name

    def eval(self, name, method, *args, **kwargs):
        """Evaluate a method of a particular object.

        Parameters
        ----------
        name : string
            Name of the object.
        method : string
            Name of the method to evaluate.
        args and kwargs: arguments
            Arguments pass to the method to evaluate.
        """
        assert name in self._objs_order and isinstance(method, str)
        # Eval the method :
        exec('self[name].' + method + '(*args, **kwargs)')

    # ----------- PARENT -----------
    @property
    def parent(self):
        """Get the parent value."""
        return self._cnode.parent

    @parent.setter
    def parent(self, value):
        """Set parent value."""
        self._cnode.parent = value

    # ----------- VISIBLE_OBJ -----------
    @property
    def visible_obj(self):
        """Get the visible_obj value."""
        return self._visible_obj

    @visible_obj.setter
    def visible_obj(self, value):
        """Set visible_obj value."""
        assert isinstance(value, bool)
        self._cnode.visible = value
        self._visible_obj = value



"""Load sleep files.

This file contain functions to load :
- European Data Format (*.edf)
- Micromed (*.trc)
- BrainVision (*.vhdr)
- ELAN (*.eeg)
- Hypnogram (*.hyp)
"""
import os
import io
import numpy as np
import datetime
from warnings import warn
import logging

from .dependencies import is_mne_installed
from .dialog import dialog_load
from .mneio import mne_switch
from .rw_hypno import (read_hypno, oversample_hypno)
from .rw_utils import get_file_ext
from .write_data import write_csv
from ..config import PROFILER
from ..io import merge_annotations
from ..utils import get_dsf, vispy_array, sleepstats

logger = logging.getLogger('visbrain')

__all__ = ['ReadSleepData', 'get_sleep_stats']


class ReadSleepData(object):
    """Main class for reading sleep data."""

    def __init__(self, data, channels, sf, hypno, href, preload, use_mne,
                 downsample, kwargs_mne, annotations):
        """Init."""
        # ========================== LOAD DATA ==========================
        # Dialog window if data is None :
        if data is None:
            data = dialog_load(self, "Open dataset", '',
                               "BrainVision (*.vhdr);;EDF (*.edf);;"
                               "GDF (*.gdf);;BDF (*.bdf);;Elan (*.eeg);;"
                               "EGI (*.egi);;MFF (*.mff);;CNT (*.cnt);;"
                               "Micromed (*.trc);;EEGLab (*.set);;REC (*.rec)")
            upath = os.path.split(data)[0]
        else:
            upath = ''

        if isinstance(data, str):  # file is defined
            # ---------- USE SLEEP or MNE ----------
            # Find file extension :
            file, ext = get_file_ext(data)
            # Force to use MNE if preload is False :
            use_mne = True if not preload else use_mne
            # Get if the file has to be loaded using Sleep or MNE python :
            sleep_ext = ['.eeg', '.vhdr', '.edf', '.trc', '.rec']
            use_mne = True if ext not in sleep_ext else use_mne

            if use_mne:
                is_mne_installed(raise_error=True)

            # ---------- LOAD THE FILE ----------
            if use_mne:  # Load using MNE functions
                logger.debug("Load file using MNE-python")
                kwargs_mne['preload'] = preload
                args = mne_switch(file, ext, downsample, **kwargs_mne)
            else:  # Load using Sleep functions
                logger.debug("Load file using Sleep")
                args = sleep_switch(file, ext, downsample)
            # Get output arguments :
            (sf, downsample, dsf, data, channels, n, offset, annot) = args
            info = ("Data successfully loaded (%s):"
                    "\n- Sampling-frequency : %.2fHz"
                    "\n- Number of time points (before down-sampling): %i"
                    "\n- Down-sampling frequency : %.2fHz"
                    "\n- Number of time points (after down-sampling): %i"
                    "\n- Number of channels : %i"
                    )
            n_channels, n_pts_after = data.shape
            logger.info(info % (file + ext, sf, n, downsample, n_pts_after,
                                n_channels))
            PROFILER("Data file loaded", level=1)

        elif isinstance(data, np.ndarray):  # array of data is defined
            if not isinstance(sf, (int, float)):
                raise ValueError("When passing raw data, the sampling "
                                 "frequency parameter, sf, must either be an "
                                 "integer or a float.")
            file = annot = None
            offset = datetime.time(0, 0, 0)
            dsf, downsample = get_dsf(downsample, sf)
            n = data.shape[1]
            data = data[:, ::dsf]
        else:
            raise IOError("The data should either be a string which refer to "
                          "the path of a file or an array of raw data of shape"
                          " (n_electrodes, n_time_points).")

        # Keep variables :
        self._file = file
        self._annot_file = np.c_[merge_annotations(annotations, annot)]
        self._N = n
        self._dsf = dsf
        self._sfori = float(sf)
        self._toffset = offset.hour * 3600. + offset.minute * 60. + \
            offset.second
        time = np.arange(n)[::dsf] / sf
        self._sf = float(downsample) if downsample is not None else float(sf)

        # ========================== LOAD HYPNOGRAM ==========================
        # Dialog window for hypnogram :
        if hypno is None:
            hypno = dialog_load(self, "Open hypnogram", upath,
                                "Text file (*.txt);;Elan (*.hyp);;"
                                "CSV file (*.csv);;EDF+ file(*.edf);"
                                ";All files (*.*)")
            hypno = None if hypno == '' else hypno
        if isinstance(hypno, np.ndarray):  # array_like
            if len(hypno) == n:
                hypno = hypno[::dsf]
            else:
                raise ValueError("Then length of the hypnogram must be the "
                                 "same as raw data")
        if isinstance(hypno, str):  # (*.hyp / *.txt / *.csv)
            hypno, _ = read_hypno(hypno, time=time, datafile=file)
            # Oversample then downsample :
            hypno = oversample_hypno(hypno, self._N)[::dsf]
            PROFILER("Hypnogram file loaded", level=1)

        # ========================== CHECKING ==========================
        # ---------- DATA ----------
        # Check data shape :
        if data.ndim != 2:
            raise ValueError("The data must be a 2D array")
        nchan, npts = data.shape

        # ---------- CHANNELS ----------
        if (channels is None) or (len(channels) != nchan):
            warn("The number of channels must be " + str(nchan) + ". Default "
                 "channel names will be used instead.")
            channels = ['chan' + str(k) for k in range(nchan)]
        # Clean channel names :
        patterns = ['eeg', 'EEG', 'ref']
        chanc = []
        for c in channels:
            # Remove informations after . :
            c = c.split('.')[0]
            c = c.split('-')[0]
            # Exclude patterns :
            for i in patterns:
                c = c.replace(i, '')
            # Remove space :
            c = c.replace(' ', '')
            c = c.strip()
            chanc.append(c)

        # ---------- STAGE ORDER ----------
        # href checking :
        absref = ['art', 'wake', 'n1', 'n2', 'n3', 'rem']
        absint = [-1, 0, 1, 2, 3, 4]
        if href is None:
            href = absref
        elif (href is not None) and isinstance(href, list):
            # Force lower case :
            href = [k.lower() for k in href]
            # Check that all stage are present :
            for k in absref:
                if k not in href:
                    raise ValueError(k + " not found in href.")
            # Force capitalize :
            href = [k.capitalize() for k in href]
            href[href.index('Rem')] = 'REM'
        else:
            raise ValueError("The href parameter must be a list of string and"
                             " must contain 'art', 'wake', 'n1', 'n2', 'n3' "
                             "and 'rem'")
        # Conversion variable :
        absref = ['Art', 'Wake', 'N1', 'N2', 'N3', 'REM']
        conv = {absint[absref.index(k)]: absint[i] for i, k in enumerate(href)}

        # ---------- HYPNOGRAM ----------
        if hypno is None:
            hypno = np.zeros((npts,), dtype=np.float32)
        else:
            n = len(hypno)
            # Check hypno values :
            if (hypno.min() < -1.) or (hypno.max() > 4) or (n != npts):
                warn("\nHypnogram values must be comprised between -1 and 4 "
                     "(see Iber et al. 2007). Use:\n-1 -> Art (optional)\n 0 "
                     "-> Wake\n 1 -> N1\n 2 -> N2\n 3 -> N4\n 4 -> REM\nEmpty "
                     "hypnogram will be used instead")
                hypno = np.zeros((npts,), dtype=np.float32)

        # ---------- SCALING ----------
        # Check amplitude of the data and if necessary apply re-scaling
        if np.abs(np.ptp(data, 0).mean()) < 0.1:
            warn("Wrong data amplitude for Sleep software.")
            data *= 1e6

        # ---------- CONVERSION ----------=
        # Convert data and hypno to be contiguous and float 32 (for vispy):
        self._data = vispy_array(data)
        self._hypno = vispy_array(hypno)
        self._time = vispy_array(time)
        self._channels = chanc
        self._href = href
        self._hconv = conv
        PROFILER("Check data", level=1)


def sleep_switch(file, ext, downsample):
    """Switch between sleep data files.

    Parameters
    ----------
    file : string
        Path to the file to load.
    ext : string
        Extension name (e.g. '.eeg')
    downsample : int
        Down-sampling frequency.

    Returns
    -------
    sf : float
        The original sampling-frequency.
    downsample : float
        The down-sampling frequency used.
    dsf : int
        The down-sampling factor.
    data : array_like
        The raw data of shape (n_channels, n_points)
    channels : list
        List of channel names.
    n : int
        Number of time points before down-sampling.
    start_time : datetime.time
        The time offset.
    annotations : array_like
        Array of annotations.
    """
    # Get full path :
    path = file + ext

    if ext == '.vhdr':  # BrainVision
        return read_bva(path, downsample)

    if ext == '.eeg':  # Elan
        return read_elan(path, downsample)

    elif ext in ['.edf', '.rec']:  # European Data Format
        return read_edf(path, downsample)

    elif ext == '.trc':  # Micromed
        return read_trc(path, downsample)

    else:  # None
        raise ValueError("*" + ext + " files are currently not supported.")


###############################################################################
###############################################################################
#                               LOAD FILES
###############################################################################
###############################################################################

def read_edf(path, downsample):
    """Read data from a European Data Format (edf) file.

    Use phypno class for reading EDF files:
        http: // phypno.readthedocs.io / api / phypno.ioeeg.edf.html

    Parameters
    ----------
    path: str
        Filename(with full path) to EDF file
    downsample : int
        Down-sampling frequency.

    Returns
    -------
    sf : int
        The sampling frequency.
    data : array_like
        The data organised as well(n_channels, n_points)
    chan : list
        The list of channel's names.
    n : int
        Number of points in the original data
    start_time : array_like
        Starting time of the recording (hh:mm:ss)
    annotations : array_like
        Array of annotations.
    """
    assert os.path.isfile(path)

    from ..utils.sleep.edf import Edf

    edf = Edf(path)

    # Return header informations
    _, start_time, sf, chan, n_samples, _ = edf.return_hdr()
    start_time = start_time.time()

    # Keep only data channels (e.g excludes marker chan)
    freqs = np.unique(edf.hdr['n_samples_per_record']) / edf.hdr[
        'record_length']
    sf = freqs.max()

    if len(freqs) != 1:
        bad_chans = np.where(edf.hdr['n_samples_per_record'] < sf)
        chan = np.delete(chan, bad_chans)

    # Load all samples of selected channels
    np.seterr(divide='ignore', invalid='ignore')
    data = edf.return_dat(chan, 0, n_samples)

    # Get original signal length :
    n = data.shape[1]

    # Get down-sample factor :
    sf = float(sf)
    chan = list(chan)
    dsf, downsample = get_dsf(downsample, sf)

    return sf, downsample, dsf, data[:, ::dsf], chan, n, start_time, None


def read_trc(path, downsample):
    """Read data from a Micromed (trc) file (version 4).

    Poor man's version of micromedio.py from Neo package
    (https://pythonhosted.org/neo/)

    Parameters
    ----------
    path : str
        Filename(with full path) to .trc file
    downsample : int
        Down-sampling frequency.

    Returns
    -------
    sf : float
        The sampling frequency.
    downsample : float
        The downsampling frequency
    data : array_like
        The data organised as well(n_channels, n_points)
    chan : list
        The list of channel's names.
    n : int
        Number of samples before down-sampling.
    start_time : array_like
        Starting time of the recording (hh:mm:ss)
    annotations : array_like
        Array of annotations.
    """
    import struct

    def read_f(f, fmt):
        return struct.unpack(fmt, f.read(struct.calcsize(fmt)))

    with io.open(path, 'rb') as f:
        # Read header
        f.seek(175, 0)
        header_version, = read_f(f, 'b')
        assert header_version == 4

        f.seek(138, 0)
        data_start_offset, n_chan, _, sf, nbytes = read_f(f, 'IHHHH')

        f.seek(128, 0)
        day, month, year, hour, minute, sec = read_f(f, 'bbbbbb')
        start_time = datetime.time(hour, minute, sec)

        # Raw data
        f.seek(data_start_offset, 0)
        m_raw = np.fromstring(f.read(), dtype='u' + str(nbytes))
        m_raw = m_raw.reshape((int(m_raw.size / n_chan), n_chan)).transpose()

        # Read label / gain
        gain = []
        chan = []
        logical_ground = []
        data = np.empty(shape=m_raw.shape, dtype=np.float32)

        f.seek(176, 0)
        zone_names = ['ORDER', 'LABCOD']
        zones = {}
        for zname in zone_names:
            zname2, pos, length = read_f(f, '8sII')
            zones[zname] = zname2, pos, length

        zname2, pos, length = zones['ORDER']
        f.seek(pos, 0)
        code = np.fromfile(f, dtype='u2', count=n_chan)

        for c in range(n_chan):
            zname2, pos, length = zones['LABCOD']
            f.seek(pos + code[c] * 128 + 2, 0)

            chan = np.append(chan, f.read(6).decode('utf-8').strip())
            logical_min, logical_max, logic_ground_chan, physical_min, \
                physical_max = read_f(f, 'iiiii')

            logical_ground = np.append(logical_ground, logic_ground_chan)

            gain = np.append(gain, float(physical_max - physical_min) /
                             float(logical_max - logical_min + 1))

    # Multiply by gain
    m_raw = m_raw - logical_ground[:, np.newaxis]
    data = m_raw * gain[:, np.newaxis].astype(np.float32)

    # Get original signal length :
    n = data.shape[1]

    # Get down-sample factor :
    sf = float(sf)
    chan = list(chan)
    dsf, downsample = get_dsf(downsample, sf)

    return sf, downsample, dsf, data[:, ::dsf], chan, n, start_time, None


def read_bva(path, downsample, read_markers=False):
    """Read data from a BrainVision (*.vhdr) file.

    Poor man's version of https: // gist.github.com / breuderink / 6266871

    Assumes that data are saved with the following parameters:
        - Data format: Binary
        - Orientation: Multiplexed
        - Format: int16

    Parameters
    ----------
    path : str
        Filename(with full path) to .vhdr file. Data file must be in the
        same directory.
    downsample : int
        Down-sampling frequency.
    read_markers : bool | False
        Import markers from the .vmrk files as annotations

    Returns
    -------
    sf : float
        The sampling frequency.
    data : array_like
        The data organised as well(n_channels, n_points)
    chan : list
        The list of channel's names.
    n : int
        Number of points before down-sampling.
    start_time : array_like
        Starting time of the recording (hh:mm:ss)
    annotations : array_like
        Array of annotations.
    """
    import re

    assert os.path.isfile(path)

    # Read header
    ent = np.genfromtxt(path, delimiter='\n', usecols=[0],
                        dtype=None, skip_header=0, encoding='utf-8')

    for item in ent:
        if 'DataFile=' in item:
            data_file = item.split('=')[1]
            data_path = os.path.join(os.path.dirname(path), data_file)
            assert os.path.isfile(data_path)
        elif 'MarkerFile=' in item:
            marker_file = item.split('=')[1]
            marker_path = os.path.join(os.path.dirname(path), marker_file)
        elif 'NumberOfChannels=' in item:
            n_chan = int(re.findall(r'\d+', item)[0])
        elif 'SamplingInterval=' in item:
            si = float(re.findall(r"[-+]?\d*\.\d+|\d+", item)[0])
            sf = 1 / (si * 0.000001)
        elif 'DataFormat' in item:
            data_format = item.split('=')[1]
        elif 'BinaryFormat' in item:
            binary_format = item.split('=')[1]
        elif 'DataOrientation' in item:
            data_orient = item.split('=')[1]

    # Check binary format
    assert "BINARY" in data_format
    assert "INT_16" in binary_format
    assert "MULTIPLEXED" in data_orient

    # Extract channel labels and resolution
    start_label = np.array(np.where(np.char.find(ent, 'Ch1=') == 0)).min()
    chan = {}
    resolution = np.empty(shape=n_chan)

    for i, j in enumerate(range(start_label, start_label + n_chan)):
        chan[i] = re.split(r'\W+', ent[j])[1]
        resolution[i] = float(ent[j].split(",")[2])

    chan = np.array(list(chan.values())).flatten()

    # Read marker file (if present) to extract recording time
    if os.path.isfile(marker_path):
        vmrk = np.genfromtxt(marker_path, delimiter='\n', usecols=[0],
                             dtype=None, skip_header=0, encoding='utf-8')

        # Read start-time
        for item in vmrk:
            if 'New Segment' in item:
                st = re.split(r'\W+', item)[-1]
                start_time = datetime.time(int(st[8:10]), int(st[10:12]),
                                           int(st[12:14]))
                break
            else:
                start_time = datetime.time(0, 0, 0)

        # Read markers
        if read_markers:
            onsets = np.array([], dtype=float)
            durations = np.array([], dtype=float)
            descriptions = np.array([], dtype=str)
            for item in vmrk:
                if 'Mk' in item and ';' not in item:
                    onsets = np.append(onsets, int(
                        re.sub(r'\s', '', item).split(',')[2]))
                    durations = np.append(durations, int(
                        re.sub(r'\s', '', item).split(',')[3]))
                    descriptions = np.append(descriptions, re.sub(
                        r'\s', '', item).split(',')[1])
                    anot = np.c_[onsets, durations, descriptions]
        else:
            anot = None

    with io.open(data_path, 'rb') as f:
        raw = f.read()
        size = int(len(raw) / 2)

        ints = np.ndarray((n_chan, int(size / n_chan)),
                          dtype='<i2', order='F', buffer=raw)

        data = np.float32(np.diag(resolution)).dot(ints)

    # Get original signal length :
    n = data.shape[1]

    # Get down-sample factor :
    sf = float(sf)
    chan = list(chan)
    dsf, downsample = get_dsf(downsample, sf)

    return sf, downsample, dsf, data[:, ::dsf], chan, n, start_time, anot


def read_elan(path, downsample):
    """Read data from a ELAN (eeg) file.

    Elan format specs: http: // elan.lyon.inserm.fr/

    Parameters
    ----------
    path : str
        Filename(with full path) to Elan .eeg file
    downsample : int
        Down-sampling frequency.

    Returns
    -------
    sf : int
        The sampling frequency.
    data : array_like
        The data organised as well(n_channels, n_points)
    chan : list
        The list of channel's names.
    n : int
        Number of samples before down-sampling.
    start_time : array_like
        Starting time of the recording (hh:mm:ss)
    annotations : array_like
        Array of annotations.
    """
    header = path + '.ent'

    assert os.path.isfile(path)
    assert os.path.isfile(header)

    # Read .ent file
    ent = np.genfromtxt(header, delimiter='\n', usecols=[0],
                        dtype=None, skip_header=0, encoding='utf-8')

    # eeg file version
    eeg_version = ent[0]

    if eeg_version == 'V2':
        nb_oct = 2
        formread = '>i2'
    elif eeg_version == 'V3':
        nb_oct = 4
        formread = '>i4'

    # Sampling rate
    sf = 1. / float(ent[8])

    # Record starting time
    if ent[4] != "No time":
        hour, minutes, sec = ent[4].split(':')
        start_time = datetime.time(int(hour), int(minutes), int(sec))
        day, month, year = ent[3].split(':')
    else:
        start_time = datetime.time(0, 0, 0)

    # Channels
    # TODO: changed from np.int to int
    nb_chan = int(ent[9])
    nb_chan = nb_chan

    # Last 2 channels do not contain data
    nb_chan_data = nb_chan - 2
    chan_list = slice(nb_chan_data)
    chan = ent[10:10 + nb_chan_data]

    # Gain
    gain = np.zeros(nb_chan)
    offset1 = 9 + 3 * nb_chan
    offset2 = 9 + 4 * nb_chan
    offset3 = 9 + 5 * nb_chan
    offset4 = 9 + 6 * nb_chan

    for i in np.arange(1, nb_chan + 1):

        min_an = float(ent[offset1 + i])
        max_an = float(ent[offset2 + i])
        min_num = float(ent[offset3 + i])
        max_num = float(ent[offset4 + i])

        gain[i - 1] = (max_an - min_an) / (max_num - min_num)
    if gain.dtype != np.float32:
        gain = gain.astype(np.float32, copy=False)

    # Load memmap
    nb_bytes = os.path.getsize(path)
    nb_samples = int(nb_bytes / (nb_oct * nb_chan))

    # TODO: bugfix: order should be string, not set
    # ie changed from {'F'} to 'F'
    m_raw = np.memmap(path, dtype=formread, mode='r',
                      shape=(nb_chan, nb_samples), order='F')

    # Get original signal length :
    n = m_raw.shape[1]

    # Get down-sample factor :
    sf = float(sf)
    chan = list(chan)
    dsf, downsample = get_dsf(downsample, sf)

    # Multiply by gain :
    data = m_raw[chan_list, ::dsf] * \
        gain[chan_list][..., np.newaxis]

    return sf, downsample, dsf, data, chan, n, start_time, None


def get_sleep_stats(hypno_file, output_file=None):
    """Compute sleep statistics from hypnogram file and export them in csv.

    Sleep statistics specifications:

        * Time in Bed (TIB) : total duration of the hypnogram.
        * Total Dark Time (TDT) : duration of the hypnogram from beginning
          to last period of sleep.
        * Sleep Period Time (SPT) : duration from first to last period of
          sleep.
        * Wake After Sleep Onset (WASO) : duration of wake periods within SPT
        * Sleep Efficiency (SE) : TST / TDT * 100 (%).
        * Total Sleep Time (TST) : SPT - WASO.
        * W, N1, N2, N3 and REM: sleep stages duration.
        * % (W, ... REM) : sleep stages duration expressed in percentages of
          TDT.
        * Latencies: latencies of sleep stages from the beginning of the
          record.

    (All values except SE and percentages are expressed in minutes)

    Parameters
    ----------
    hypno_file : string
        Full path to the hypnogram file.
    output_file : string | None
        Full path to the output file. If no file is provided, sleep statictics
        are print out to the terminal.
    """
    # File conversion :
    if output_file is not None:  # Check extension
        ext = os.path.splitext(output_file)[1][1:].strip().lower()
        if ext == '':
            output_file = output_file + '.csv'

    # Load hypnogram
    hypno, sf_hyp = read_hypno(hypno_file)
    if sf_hyp < 1:
        mult = int(np.round(len(hypno) / sf_hyp))
        hypno = oversample_hypno(hypno, mult)
        sf_hyp = 1

    # Get sleep stats
    stats = sleepstats(hypno, sf_hyp=sf_hyp)
    stats['File'] = hypno_file
    print('\nSLEEP STATS\n===========')
    keys, val = [''] * len(stats), [''] * len(stats)
    # Fill table :
    for num, (k, v) in enumerate(stats.items()):
        print(k, '\t', str(v))
        # Remember variables :
        keys[int(num)] = k
        val[int(num)] = str(v)
    if output_file is not None:
        write_csv(output_file, zip(keys, val))
        print('===========\nCSV file saved to:', output_file)



"""
Hardware accelerated graphics for neuscientific data
====================================================

visbrain is an open-source Python software mainly dedicated to the
visualization of neuroscientific data. It's developped on top of VisPy which
provides graphic renderings offloaded to the GPU.

Right now, visbrain contains five modules :
* Brain : visualize EEG/MEG/Intracranial data and connectivity in a standard
  MNI 3D brain.
* Sleep : visualize polysomnographic data and hypnogram edition.
* Signal : data mining module for signal inspection.
* Figure : figure-layout for high-quality publication-like figures.
* Topo : topographic representations

See http://visbrain.org/ for a complete and step-by step documentation
"""
import sys as _sys

__version__ = "0.4.3"


# PyQt5 crash if an error occured. This small function fix it for all modules
# to retrieve the PyQt4 behavior :


def _pyqt4_behavior(type, value, tback):
    """Retrieve PyQt4 behavior if an error occured."""
    _sys.__excepthook__(type, value, tback)


_sys.excepthook = _pyqt4_behavior



"""Display signals into a grid..

A large portion of this code was taken from the example developped by the vispy
team :
https://github.com/vispy/vispy/blob/master/examples/demo/gloo/realtime_signals.py
"""
import numpy as np

from vispy import gloo, visuals
from vispy.visuals.shaders import Function
from vispy.scene.visuals import create_visual_node

from sleep_scoring.utils import vispy_array, wrap_properties, color2vb, transient
# from sleep_scoring.io import is_opengl_installed


__all__ = ('Hypnogram')


STAGES = ['art', 'wake', 'rem', 'n1', 'n2', 'n3']

vec2to4 = Function("""
    vec4 vec2to4(vec2 inp) {
        return vec4(inp, 0, 1);
    }
""")

vertex_shader = """
#version 120
varying vec4 pos;
varying vec4 v_color;
float pos_y;

void main() {

    // Position conversion :
    pos = $to_vec4($position);
    if (pos.y == $u_art){ // ART
        pos.y = $u_art_visual;
        v_color = $u_art_color;
    }
    else if (pos.y == $u_wake){ // WAKE
        pos.y = $u_wake_visual;
        v_color = $u_wake_color;
    }
    else if (pos.y == $u_rem){ // REM
        pos.y = $u_rem_visual;
        v_color = $u_rem_color;
    }
    else if (pos.y == $u_n1){ // N1
        pos.y = $u_n1_visual;
        v_color = $u_n1_color;
    }
    else if (pos.y == $u_n2){ // N2
        pos.y = $u_n2_visual;
        v_color = $u_n2_color;
    }
    else if (pos.y == $u_n3){ // N3
        pos.y = $u_n3_visual;
        v_color = $u_n3_color;
    }

    // Color conversion :
    if ($transient != 10.){
        v_color = vec4(.95, .95, .95, 1.);
    }

    // Uniform color :
    if ($u_unicolor == 1){
        v_color = vec4(0., 0., 0., 1.);
    }

    gl_Position = $transform(pos);
}
"""

fragment_shader = """
#version 120
varying vec4 v_color;

void main() {
    gl_FragColor = v_color;
}
"""


class HypogramVisual(visuals.Visual):
    """Visual class for grid of signals.

    Parameters
    ----------
    data : array_like
        Array of data of shape (n_pts,).
    time : array_like | None
        Array of time points of shape (n_pts,)
    art, wake, rem, n1, n2, n3 :
        Stage identification inside the data array.
    art_visual, wake_visual, rem_visual, n1_visual, n2_visual, n3_visual :
        Stage order when plotting.
    art_color, wake_color, rem_color, n1_color, n2_color, n3_color :
        Stage color.
    line_width : float | 2.
        Line with of the hypnogram.
    antialias : bool | False
        Use anti-aliasing line.
    unicolor : bool | False
        Use a uni black color for the hypnogram.
    """

    def __len__(self):
        """Return the number of time points."""
        return self._n

    def __init__(self, data, time=None, art=-1, wake=0, n1=1, n2=2, n3=3,
                 rem=4, art_visual=1, wake_visual=0, rem_visual=-1,
                 n1_visual=-2, n2_visual=-3, n3_visual=-4, art_color='#8bbf56',
                 wake_color='#56bf8b', rem_color='#bf5656', n1_color='#aabcce',
                 n2_color='#405c79', n3_color='#0b1c2c', line_width=2.,
                 antialias=False, unicolor=False):
        """Init."""
        # =========================== VISUALS ===========================
        visuals.Visual.__init__(self, vertex_shader, fragment_shader)
        self._draw_mode = 'line_strip'

        # =========================== BUFFERS ===========================
        pos_2 = np.random.rand(1, 2).astype(np.float32)
        pos_1 = np.random.rand(1).astype(np.float32)
        self._position_vbo = gloo.VertexBuffer(pos_2)
        self._transient_vbo = gloo.VertexBuffer(pos_1)
        self.shared_program.vert['position'] = self._position_vbo
        self.shared_program.vert['transient'] = self._transient_vbo
        self._program.vert['to_vec4'] = vec2to4
        self.line_width = line_width
        self.antialias = antialias
        self.unicolor = unicolor
        # Stage :
        self.art = art
        self.wake = wake
        self.rem = rem
        self.n1 = n1
        self.n2 = n2
        self.n3 = n3
        # Stage visual :
        self.art_visual = art_visual
        self.wake_visual = wake_visual
        self.rem_visual = rem_visual
        self.n1_visual = n1_visual
        self.n2_visual = n2_visual
        self.n3_visual = n3_visual
        # Stage color :
        self.art_color = art_color
        self.wake_color = wake_color
        self.rem_color = rem_color
        self.n1_color = n1_color
        self.n2_color = n2_color
        self.n3_color = n3_color

        # Set data :
        self.set_data(data, time)
        self.freeze()

    def set_data(self, data, time=None):
        """Set data to the grid of signals.

        Parameters
        ----------
        data : array_like
            Array of data of shape (n_pts,).
        time : array_like | None
            Array of time points of shape (n_pts,)
        """
        data = np.asarray(data)
        assert data.ndim == 1
        self._n = len(data)
        time = np.arange(len(data)) if time is None else np.asarray(time)
        assert len(time) == len(self)

        # Transient detection :
        self.transient = data

        #
        self._pos = vispy_array(np.c_[time, data])
        self._position_vbo.set_data(self._pos)

    def set_stage(self, stage, idx_start, idx_end):
        """Set stage.

        Parameters
        ----------
        stage : str, int
            Stage to define. Should either be a string (e.g 'art', 'rem'...) or
            an integer.
        idx_start : int
            Index where the stage begin.
        idx_end : int
            Index where the stage finish.
        """
        if isinstance(stage, str):
            assert stage in STAGES
            stage = eval('self.%s' % stage)
        self._pos[idx_start:idx_end, 1] = stage
        self.transient = self._pos[:, 1]

    def _prepare_transforms(self, view):
        """Call for the first rendering."""
        tr = view.transforms
        view_vert = view.view_program.vert
        view_vert['transform'] = tr.get_transform()

    def _prepare_draw(self, view=None):
        """Function called everytime there's a camera update."""
        try:
            import OpenGL.GL as GL
            GL.glLineWidth(self._line_width)
            if self._antialias:
                GL.glEnable(GL.GL_LINE_SMOOTH)
            else:
                GL.glDisable(GL.GL_LINE_SMOOTH)
        except:
            pass

    def min_visual(self):
        return min(self.art_visual, self.wake_visual, self.rem_visual,
                   self.n1_visual, self.n2_visual, self.n3_visual)

    def max_visual(self):
        return max(self.art_visual, self.wake_visual, self.rem_visual,
                   self.n1_visual, self.n2_visual, self.n3_visual)

    # ========================================================================
    # ========================================================================
    # PROPERTIES
    # ========================================================================
    # ========================================================================
    # ----------- TIME -----------
    @property
    def time(self):
        """Get the time value."""
        return self._pos[:, 0]

    # ----------- DATA -----------
    @property
    def data(self):
        """Get the data value."""
        return self._pos[:, 1]

    # ----------- ART -----------
    @property
    def art(self):
        """Get the art value."""
        return self._art

    @art.setter
    @wrap_properties
    def art(self, value):
        """Set art value."""
        assert isinstance(value, (int, float))
        self.shared_program.vert['u_art'] = value
        self._art = value

    # ----------- WAKE -----------
    @property
    def wake(self):
        """Get the wake value."""
        return self._wake

    @wake.setter
    @wrap_properties
    def wake(self, value):
        """Set wake value."""
        assert isinstance(value, (int, float))
        self.shared_program.vert['u_wake'] = value
        self._wake = value

    # ----------- REM -----------
    @property
    def rem(self):
        """Get the rem value."""
        return self._rem

    @rem.setter
    @wrap_properties
    def rem(self, value):
        """Set rem value."""
        assert isinstance(value, (int, float))
        self.shared_program.vert['u_rem'] = value
        self._rem = value

    # ----------- N1 -----------
    @property
    def n1(self):
        """Get the n1 value."""
        return self._n1

    @n1.setter
    @wrap_properties
    def n1(self, value):
        """Set n1 value."""
        assert isinstance(value, (int, float))
        self.shared_program.vert['u_n1'] = value
        self._n1 = value

    # ----------- N2 -----------
    @property
    def n2(self):
        """Get the n2 value."""
        return self._n2

    @n2.setter
    @wrap_properties
    def n2(self, value):
        """Set n2 value."""
        assert isinstance(value, (int, float))
        self.shared_program.vert['u_n2'] = value
        self._n2 = value

    # ----------- N3 -----------
    @property
    def n3(self):
        """Get the n3 value."""
        return self._n3

    @n3.setter
    @wrap_properties
    def n3(self, value):
        """Set n3 value."""
        assert isinstance(value, (int, float))
        self.shared_program.vert['u_n3'] = value
        self._n3 = value

    # ----------- ART_VISUAL -----------
    @property
    def art_visual(self):
        """Get the art_visual value."""
        return self._art_visual

    @art_visual.setter
    @wrap_properties
    def art_visual(self, value):
        """Set art_visual value."""
        assert isinstance(value, (int, float))
        self.shared_program.vert['u_art_visual'] = value
        self._art_visual = value

    # ----------- WAKE_VISUAL -----------
    @property
    def wake_visual(self):
        """Get the wake_visual value."""
        return self._wake_visual

    @wake_visual.setter
    @wrap_properties
    def wake_visual(self, value):
        """Set wake_visual value."""
        assert isinstance(value, (int, float))
        self.shared_program.vert['u_wake_visual'] = value
        self._wake_visual = value

    # ----------- REM_VISUAL -----------
    @property
    def rem_visual(self):
        """Get the rem_visual value."""
        return self._rem_visual

    @rem_visual.setter
    @wrap_properties
    def rem_visual(self, value):
        """Set rem_visual value."""
        assert isinstance(value, (int, float))
        self.shared_program.vert['u_rem_visual'] = value
        self._rem_visual = value

    # ----------- N1_VISUAL -----------
    @property
    def n1_visual(self):
        """Get the n1_visual value."""
        return self._n1_visual

    @n1_visual.setter
    @wrap_properties
    def n1_visual(self, value):
        """Set n1_visual value."""
        assert isinstance(value, (int, float))
        self.shared_program.vert['u_n1_visual'] = value
        self._n1_visual = value

    # ----------- N2_VISUAL -----------
    @property
    def n2_visual(self):
        """Get the n2_visual value."""
        return self._n2_visual

    @n2_visual.setter
    @wrap_properties
    def n2_visual(self, value):
        """Set n2_visual value."""
        assert isinstance(value, (int, float))
        self.shared_program.vert['u_n2_visual'] = value
        self._n2_visual = value

    # ----------- N3_VISUAL -----------
    @property
    def n3_visual(self):
        """Get the n3_visual value."""
        return self._n3_visual

    @n3_visual.setter
    @wrap_properties
    def n3_visual(self, value):
        """Set n3_visual value."""
        assert isinstance(value, (int, float))
        self.shared_program.vert['u_n3_visual'] = value
        self._n3_visual = value

    # ----------- ART_COLOR -----------
    @property
    def art_color(self):
        """Get the art_color value."""
        return self._art_color

    @art_color.setter
    @wrap_properties
    def art_color(self, value):
        """Set art_color value."""
        color = np.squeeze(color2vb(value))
        self.shared_program.vert['u_art_color'] = color
        self._art_color = color

    # ----------- WAKE_COLOR -----------
    @property
    def wake_color(self):
        """Get the wake_color value."""
        return self._wake_color

    @wake_color.setter
    @wrap_properties
    def wake_color(self, value):
        """Set wake_color value."""
        color = np.squeeze(color2vb(value))
        self.shared_program.vert['u_wake_color'] = color
        self._wake_color = color

    # ----------- REM_COLOR -----------
    @property
    def rem_color(self):
        """Get the rem_color value."""
        return self._rem_color

    @rem_color.setter
    @wrap_properties
    def rem_color(self, value):
        """Set rem_color value."""
        color = np.squeeze(color2vb(value))
        self.shared_program.vert['u_rem_color'] = color
        self._rem_color = color

    # ----------- N1_COLOR -----------
    @property
    def n1_color(self):
        """Get the n1_color value."""
        return self._n1_color

    @n1_color.setter
    @wrap_properties
    def n1_color(self, value):
        """Set n1_color value."""
        color = np.squeeze(color2vb(value))
        self.shared_program.vert['u_n1_color'] = color
        self._n1_color = color

    # ----------- N2_COLOR -----------
    @property
    def n2_color(self):
        """Get the n2_color value."""
        return self._n2_color

    @n2_color.setter
    @wrap_properties
    def n2_color(self, value):
        """Set n2_color value."""
        color = np.squeeze(color2vb(value))
        self.shared_program.vert['u_n2_color'] = color
        self._n2_color = color

    # ----------- N3_COLOR -----------
    @property
    def n3_color(self):
        """Get the n3_color value."""
        return self._n3_color

    @n3_color.setter
    @wrap_properties
    def n3_color(self, value):
        """Set n3_color value."""
        color = np.squeeze(color2vb(value))
        self.shared_program.vert['u_n3_color'] = color
        self._n3_color = color

    # ----------- TRANSIENT -----------
    @property
    def transient(self):
        """Get the transient value."""
        return self._transient

    @transient.setter
    def transient(self, value):
        """Set transient value."""
        assert isinstance(value, np.ndarray) and len(value) == len(self)
        self._transient = np.full((len(self),), 10., dtype=np.float32)
        idx = transient(value)[0]
        idx_double = np.r_[idx, idx + 1]
        self._transient[idx_double] = value[idx_double]
        self._transient_vbo.set_data(self._transient)

    # ----------- LINE_WIDTH -----------
    @property
    def line_width(self):
        """Get the line_width value."""
        return self._line_width

    @line_width.setter
    @wrap_properties
    def line_width(self, value):
        """Set line_width value."""
        assert isinstance(value, (int, float))
        self._line_width = value

    # ----------- ANTIALIAS -----------
    @property
    def antialias(self):
        """Get the antialias value."""
        return self._antialias

    @antialias.setter
    def antialias(self, value):
        """Set antialias value."""
        assert isinstance(value, bool)
        self._antialias = value

    # ----------- UNICOLOR -----------
    @property
    def unicolor(self):
        """Get the unicolor value."""
        return self._unicolor

    @unicolor.setter
    def unicolor(self, value):
        """Set unicolor value."""
        assert isinstance(value, bool)
        self.shared_program.vert['u_unicolor'] = int(value)
        self._unicolor = value

Hypnogram = create_visual_node(HypogramVisual)  # noqa



"""Hypnogram related functions."""

import numpy as np

__all__ = ('transient', 'sleepstats')


def transient(data, xvec=None):
    """Perform a transient detection on hypnogram.

    Parameters
    ----------
    data : array_like
        The hypnogram data.
    xvec : array_like | None
        The time vector to use. If None, np.arange(len(data)) will be used
        instead.

    Returns
    -------
    t : array_like
        Hypnogram's transients.
    st : array_like
        Either the transient index (as type int) if xvec is None, or the
        converted version if xvec is not None.
    stages : array_like
        The stages for each segment.
    """
    # Transient detection :
    t = list(np.nonzero(np.abs(data[:-1] - data[1:]))[0])
    # Add first and last points :
    idx = np.vstack((np.array([-1] + t) + 1, np.array(t + [len(data) - 1]))).T
    # Get stages :
    stages = data[idx[:, 0]]
    # Convert (if needed) :
    if (xvec is not None) and (len(xvec) == len(data)):
        st = idx.copy().astype(float)
        st[:, 0] = xvec[idx[:, 0]]
        st[:, 1] = xvec[idx[:, 1]]
    else:
        st = idx

    return np.array(t), st, stages.astype(int)


def sleepstats(hypno, sf_hyp):
    """Compute sleep stats from an hypnogram vector.

    Sleep statistics specifications:
    ======================================================================

    All values except SE and percentages are expressed in minutes

    - Time in Bed (TIB): total duration of the hypnogram.

    - Total Dark Time (TDT): duration of the hypnogram from beginning
      to last period of sleep.

    - Sleep Period Time (SPT): duration from first to last period of sleep.

    - Wake After Sleep Onset (WASO): duration of wake periods within SPT

    - Sleep Efficiency (SE): TST / TDT * 100 (%).

    - Total Sleep Time (TST): SPT - WASO.

    - W, N1, N2, N3 and REM: sleep stages duration.

    - % (W, ... REM): sleep stages duration expressed in percentages of TDT.

    - Latencies: latencies of sleep stages from the beginning of the record.

    ======================================================================

    Parameters
    ----------
    hypno : array_like
        Hypnogram vector
    sf_hyp : float
        The sampling frequency of the hypnogram

    Returns
    -------
    stats: dict
        Sleep statistics (expressed in minutes)
    """
    stats = {}
    tov = np.nan

    # Downsample to 1 value per second
    hypno = hypno[::int(sf_hyp)]

    stats['TIB'] = len(hypno)
    stats['TDT'] = np.where(hypno != 0)[0].max() if np.nonzero(
        hypno)[0].size else tov

    # Duration of each sleep stages
    stats['Art'] = hypno[hypno == -1].size
    stats['W'] = hypno[hypno == 0].size
    stats['N1'] = hypno[hypno == 1].size
    stats['N2'] = hypno[hypno == 2].size
    stats['N3'] = hypno[hypno == 3].size
    stats['REM'] = hypno[hypno == 4].size

    # Sleep stage latencies
    stats['LatN1'] = np.where(hypno == 1)[0].min() if 1 in hypno else tov
    stats['LatN2'] = np.where(hypno == 2)[0].min() if 2 in hypno else tov
    stats['LatN3'] = np.where(hypno == 3)[0].min() if 3 in hypno else tov
    stats['LatREM'] = np.where(hypno == 4)[0].min() if 4 in hypno else tov

    if not np.isnan(stats['LatN1']) and not np.isnan(stats['TDT']):
        hypno_s = hypno[stats['LatN1']:stats['TDT']]

        stats['SPT'] = hypno_s.size
        stats['WASO'] = hypno_s[hypno_s == 0].size
        stats['TST'] = stats['SPT'] - stats['WASO']
    else:
        stats['SPT'] = tov
        stats['WASO'] = tov
        stats['TST'] = tov

    # Convert to minutes
    for key, value in stats.items():
        stats[key] = value / 60

    stats['SE'] = np.round(stats['TST'] / stats['TDT'] * 100., 2)
    stats['Units'] = 'minutes'

    return stats



"""Read annotations."""
import numpy as np
from .dependencies import is_mne_installed

__all__ = ['annotations_to_array', 'merge_annotations']


def annotations_to_array(annotations, default_txt='enter annotations'):
    """Convert annotations to array.

    Parameters
    ----------
    annotations : string, array_like, mne.io.annotations.Annotations
        Annotations to use. Should either be a string for a path to an
        annotation file, an Annotation instance from MNE or a 1-D or 3-D array
        of annotations (e.g. start, end, text).
    default_txt : string | 'enter annotations'
        Default text to use if no text is given.

    Returns
    -------
    start : array_like
        First array of float.
    end : array_like
        Second array of float.
    text : array_like
        Array of text.
    """
    if annotations is None:
        start = end = text = np.array([])
    elif isinstance(annotations, str):  # 'file.txt'
        # Get starting/ending/annotation :
        start, end, text = np.genfromtxt(annotations, delimiter=',',
                                         dtype=str, encoding='utf-8').T
    elif isinstance(annotations, (np.ndarray, list)):  # array of annotations
        annotations = np.asarray(annotations)
        if (annotations.ndim == 1):
            start = end = annotations
            text = np.array(['enter annotations'] * len(start))
        elif (annotations.ndim == 2) and (annotations.shape[1] == 3):
            start, end, text = annotations.T
            start = start.astype(float)
    elif is_mne_installed():  # MNE annotations
        import mne
        if isinstance(annotations, mne.annotations.Annotations):
            start = annotations.onset
            end = annotations.onset + annotations.duration
            text = annotations.description
    else:
        raise ValueError("Annotation's type not supported.")

    return start.astype(float), end.astype(float), text


def merge_annotations(*args):
    """Merge several annotations together.

    Parameters
    ----------
    args : string, array_like, mne.io.annotations.Annotations
        Annotation file / array / MNE instance.

    Returns
    -------
    start : array_like
        First array of float.
    end : array_like
        Second array of float.
    text : array_like
        Array of text.
    """
    start = end = text = np.array([])
    for k in args:
        # Convert annotations :
        start_t, end_t, text_t = annotations_to_array(k)
        # Extend start, end and text :
        start = np.append(start, start_t)
        end = np.append(end, end_t)
        text = np.append(text, text_t)
    return start, end, text



"""Mehods to manage interactions between the GUI and objects."""
from functools import wraps

from .gui import CbarForm
from .CbarVisual import CbarVisual
from ...utils import color2json, mpl_cmap, mpl_cmap_index, disconnect_all

__all__ = ['CbarQt']


def _run_method_if_needed(fn):
    @wraps(fn)
    def wrapper(self):
        if self._trigger_cbar:
            fn(self)
    return wrapper


def _update_obj(fn):
    @wraps(fn)
    def wrapper(self):
        if self._update_obj:
            fn(self)
    return wrapper


class CbarQt(object):
    """Link the GUI colorbar with the VisPy based colorbar.

    self[obj_name] : get the item from the GUI
    self[obj_name] = value : set the value to the VisPy based colorbar.

    Parameters
    ----------
    gui_w: PyQt widget
        The widget for adding the GUI colorar properties.
    viz_w: PyQt widget
        The widget for adding the VisPy based colorbar.
    """

    def __init__(self, gui_w, viz_w, cbobjs, parent=None, camera=None):
        """Init."""
        # --------------------------------------------------------------------
        #                          GUI COMPONENTS
        # --------------------------------------------------------------------
        # Add colorbar properties to gui_w :
        self.cbui = CbarForm()
        self.cbui.setupUi(gui_w)
        # Add VisPy based colorbar :
        self.cbviz = CbarVisual(parent=parent)
        viz_w.addWidget(self.cbviz._canvas.native)
        self.cbobjs = cbobjs
        # Add items contains in the cbobs :
        self['object'].addItems(self.cbobjs.keys())

        # --------------------------------------------------------------------
        #                         CONNECT GUI ELEMENTS
        # --------------------------------------------------------------------
        self['object'].currentIndexChanged.connect(self._fcn_change_object)
        self['bckCol'].editingFinished.connect(self._fcn_bgcolor)
        self['txtCol'].editingFinished.connect(self._fcn_text_color)
        self['ndigits'].valueChanged.connect(self._fcn_n_digits)
        self['width'].valueChanged.connect(self._fcn_width)
        self['border'].clicked.connect(self._fcn_border)
        self['bw'].valueChanged.connect(self._fcn_bw)
        self['limTxt'].clicked.connect(self._fcn_limits)
        self['ndigits'].setKeyboardTracking(False)
        self['width'].setKeyboardTracking(False)

        # _____________ COLORMAPS _____________
        self._cmaps = mpl_cmap()
        self['cmap'].addItems(self._cmaps)
        self['cmap'].currentIndexChanged.connect(self._fcn_cmap_changed)
        self['cmapRev'].clicked.connect(self._fcn_cmap_changed)

        # _____________ CLIM _____________
        self['climm'].valueChanged.connect(self._fcn_clim_changed)
        self['climM'].valueChanged.connect(self._fcn_clim_changed)
        self['climm'].setKeyboardTracking(False)
        self['climM'].setKeyboardTracking(False)

        # _____________ VMIN/VMAX _____________
        # Vmin :
        self['isvmin'].clicked.connect(self._fcn_vmin_changed)
        self['vmin'].valueChanged.connect(self._fcn_vmin_changed)
        self['under'].editingFinished.connect(self._fcn_vmin_changed)
        self['vmin'].setKeyboardTracking(False)
        # Vmax :
        self['isvmax'].clicked.connect(self._fcn_vmax_changed)
        self['vmax'].valueChanged.connect(self._fcn_vmax_changed)
        self['over'].editingFinished.connect(self._fcn_vmax_changed)
        self['vmax'].setKeyboardTracking(False)

        # _____________ TEXT _____________
        # Colorbar label :
        self['cblabel'].editingFinished.connect(self._fcn_cb_title)
        self['cbTxtSz'].valueChanged.connect(self._fcn_cb_txt_size)
        self['cbTxtSh'].valueChanged.connect(self._fcn_cb_txt_shift)
        self['cbTxtSz'].setKeyboardTracking(False)
        self['cbTxtSh'].setKeyboardTracking(False)
        # Limits label :
        self['txtSz'].valueChanged.connect(self._fcn_txt_size)
        self['txtSh'].valueChanged.connect(self._fcn_txt_shift)
        self['txtSz'].setKeyboardTracking(False)
        self['txtSh'].setKeyboardTracking(False)

        # _____________ AUTOSCALE _____________
        self['autoscale'].clicked.connect(self._fcn_cb_autoscale)
        self._trigger_cbar = True

    ###########################################################################
    ###########################################################################
    #                            GET / SET ITEMS
    ###########################################################################
    ###########################################################################
    def __getitem__(self, key):
        """Get the item from the GUI."""
        return eval('self.cbui.' + key)

    def __setitem__(self, key, value):
        """Set VisPy based colormap item from GUI properties."""
        if not isinstance(value, str):
            exec('self.cbviz.' + key + '=' + str(value))
        else:
            exec("self.cbviz." + key + "='" + value + "'")

    def __bool__(self):
        """Test if there's objects defined."""
        return bool(len(self.cbobjs.keys()))

    ###########################################################################
    ###########################################################################
    #                            LOAD / SAVE
    ###########################################################################
    ###########################################################################
    def save(self, filename):
        """Save all colorbar configurations.

        Parameters
        ----------
        filename : string
            Name of the file to be saved.
        """
        self.cbobjs.save(filename)

    def load(self, filename, **kwargs):
        """Load a colorbar configuration file.

        Parameters
        ----------
        filename : string
            Name of the file to load.
        kwargs : dict, optional, (def: {})
            Further arguments to pass to the CbarObjects class.
        """
        self.cbobjs.load(filename, **kwargs)

    def add_camera(self, camera):
        """Add a camera to the VisPy based colorbar."""
        self.cbviz._wc.camera = camera

    def to_dict(self):
        """Return a dictionary of the current visual state."""
        self.cbviz.to_dict()

    def link(self, name, fcn, minmaxfcn):
        """Link an object with a function.

        When an colormap argument changed, a function associated to the object
        is executed (only for colormap arguments i.e. cmap, clim, vmin, vmax,
        under, over).
        """
        self.cbobjs._objs[name]._fcn = fcn
        self.cbobjs._objs[name]._minmaxfcn = minmaxfcn

    def select(self, name):
        """Select an object.

        Parameters
        ----------
        name : string
            Name of the object to select.
        """
        # Get the list of all current objects :
        all_items = [self['object'].itemText(i) for i in range(
            self['object'].count())]
        if isinstance(name, str):
            idx = all_items.index(name)
        elif isinstance(name, int):
            idx = name
            name = all_items[name]
        assert name in all_items
        # Select object in the cbar object manager :
        self.cbobjs.select(name)
        self['object'].setCurrentIndex(idx)

    def setEnabled(self, name, enable=True):  # noqa
        """Deactivate an object."""
        # Get the list of all current objects :
        all_items = [self['object'].itemText(i) for i in range(
            self['object'].count())]
        if name in all_items:
            # Find the index and set it in the combobox :
            idx = all_items.index(name)
            self['object'].model().item(idx).setEnabled(enable)

    ###########################################################################
    ###########################################################################
    #                CONNECT / DISCONNECT / INITIALIZE
    ###########################################################################
    ###########################################################################
    # --------------------------------------------------------------------
    #                          INITIALIZE GUI
    # --------------------------------------------------------------------
    def _initialize(self):
        # _____________ SETTINGS _____________
        self['bckCol'].setText(str(self.cbobjs['bgcolor']))
        self['txtCol'].setText(str(self.cbobjs['txtcolor']))
        self['ndigits'].setValue(self.cbobjs['ndigits'])
        self['width'].setValue(self.cbobjs['width'])
        self['border'].setChecked(self.cbobjs['border'])
        self['bw'].setValue(self.cbobjs['bw'])
        self['limTxt'].setChecked(self.cbobjs['limtxt'])

        # _____________ COLORMAPS _____________
        idx, rev = mpl_cmap_index(self.cbobjs['cmap'])
        self['cmap'].setCurrentIndex(idx)
        self['cmapRev'].setChecked(rev)

        # _____________ CLIM _____________
        # Clim should never be None :
        self['climm'].setValue(self.cbobjs['clim'][0])
        self['climM'].setValue(self.cbobjs['clim'][1])

        # _____________ VMIN/VMAX _____________
        # Set vmin/vmax limits :
        self._check_vmin_vmax()
        # Vmin/under :
        self['isvmin'].setChecked(self.cbobjs['isvmin'])
        self['vmin'].setValue(self.cbobjs['vmin'])
        self['under'].setText(str(self.cbobjs['under']))
        # Vmax/over :
        self['isvmax'].setChecked(self.cbobjs['isvmax'])
        self['vmax'].setValue(self.cbobjs['vmax'])
        self['over'].setText(str(self.cbobjs['over']))

        # _____________ TEXT _____________
        # Colorbar :
        self['cblabel'].setText(str(self.cbobjs['cblabel']))
        self['cbTxtSz'].setValue(self.cbobjs['cbtxtsz'])
        self['cbTxtSh'].setValue(self.cbobjs['cbtxtsh'])
        # Text limits :
        self['txtSz'].setValue(self.cbobjs['txtsz'])
        self['txtSh'].setValue(self.cbobjs['txtsh'])

        # Enable to run methods to update the CbarVisual :
        self._trigger_cbar = True
        self._gui_to_visual()

    # --------------------------------------------------------------------
    #                             GUI -> VISUAL
    # --------------------------------------------------------------------
    def _gui_to_visual(self):
        # Settings :
        self._fcn_bgcolor()
        self._fcn_text_color()
        self._fcn_n_digits()
        self._fcn_width()
        self._fcn_border()
        self._fcn_bw()
        self._fcn_limits()
        # Cmap/Clim/Vmin/Vmax :
        self._fcn_cmap_changed()
        self._fcn_clim_changed()
        self._fcn_vmin_changed()
        self._fcn_vmax_changed()
        # Text :
        self._fcn_cb_title()
        self._fcn_cb_txt_size()
        self._fcn_cb_txt_shift()
        self._fcn_txt_size()
        self._fcn_txt_shift()

    ###########################################################################
    ###########################################################################
    #                              SUB-FONCTION
    ###########################################################################
    ###########################################################################
    # --------------------------------------------------------------------
    #                             SETTINGS
    # --------------------------------------------------------------------
    def _fcn_change_object(self, *args, clean=False):
        """Change colorbar object."""
        # Disconnect interactions :
        self._trigger_cbar = False
        disconnect_all(self['object'])
        # Clean drowdown box for object selection :
        if clean:
            self['object'].clear()
            self['object'].addItems(self.cbobjs.keys())
        # Get the current selected text :
        name = str(self['object'].currentText())
        # Select this object :
        self.cbobjs.select(name)
        # Update GUI :
        self._initialize()
        # Re-connect interactions :
        self._trigger_cbar = True
        self['object'].currentIndexChanged.connect(self._fcn_change_object)

    @_run_method_if_needed
    def _fcn_bgcolor(self):
        """Change background color."""
        bgcolor = color2json(self['bckCol'])
        self['bgcolor'] = bgcolor
        self.cbobjs['bgcolor'] = self.cbviz.bgcolor

    @_run_method_if_needed
    def _fcn_n_digits(self):
        """Change the number of digits."""
        ndigits = self['ndigits'].value()
        if ndigits > 0:
            self['ndigits'] = ndigits
            self.cbobjs['ndigits'] = self.cbviz.ndigits
            self['climm'].setDecimals(ndigits)
            self['climM'].setDecimals(ndigits)
            self['vmin'].setDecimals(ndigits)
            self['vmax'].setDecimals(ndigits)

    @_run_method_if_needed
    def _fcn_width(self):
        """Change colorbar width."""
        self['width'] = self['width'].value()
        self.cbobjs['width'] = self.cbviz.width

    @_run_method_if_needed
    def _fcn_border(self):
        """Set the border."""
        viz = self['border'].isChecked()
        self['border'] = viz
        self.cbobjs['border'] = self.cbviz.border
        self['bw'].setEnabled(viz)

    @_run_method_if_needed
    def _fcn_bw(self):
        """Change border width."""
        self['bw'] = self['bw'].value()
        self.cbobjs['bw'] = self.cbviz.bw

    # --------------------------------------------------------------------
    #                                CMAP
    # --------------------------------------------------------------------
    @_run_method_if_needed
    def _fcn_cmap_changed(self):
        """Change the colormap."""
        rv = self['cmapRev'].isChecked() * '_r'
        self['cmap'] = str(self['cmap'].currentText()) + rv
        self.cbobjs['cmap'] = self.cbviz.cmap
        # Run object's update function :
        self.cbobjs.update()

    # --------------------------------------------------------------------
    #                             CLIM
    # --------------------------------------------------------------------
    @_run_method_if_needed
    def _fcn_clim_changed(self):
        """Update colorbar limits."""
        # Get value (climm, climM):
        climm = float(self['climm'].value())
        clim_max = float(self['climM'].value())
        # Fix (climm, climM) limits :
        if climm < clim_max:
            # Update clim :
            self['clim'] = (climm, clim_max)
            self.cbobjs['clim'] = self.cbviz.clim
            # Update vmin/vmax limits :
            self._check_vmin_vmax()
            # Run object's update function :
            self.cbobjs.update()
        else:
            raise ValueError(str(tuple((climm, clim_max))) + " : clim.min() "
                             "> clim.max().")

    # --------------------------------------------------------------------
    #                   VMIN / VMAX / UNDER / OVER
    # --------------------------------------------------------------------
    def _check_vmin_vmax(self):
        """Activate checkboxs if vmin/vmax not None."""
        # Get clim and define step :
        climm = float(self['climm'].value())
        clim_max = float(self['climM'].value())
        step = (clim_max - climm) / 100.
        # -------------- Vmin --------------
        if not isinstance(self.cbobjs['vmin'], (int, float)):
            self.cbobjs['vmin'] = climm - 1.
        self['vmin'].setMinimum(climm)
        self['vmin'].setMaximum(clim_max)
        self['vmin'].setSingleStep(step)

        # -------------- Vmax --------------
        if not isinstance(self.cbobjs['vmax'], (int, float)):
            self.cbobjs['vmax'] = clim_max + 1.
        self['vmax'].setMinimum(climm)
        self['vmax'].setMaximum(clim_max)
        self['vmax'].setSingleStep(step)

    @_run_method_if_needed
    def _fcn_vmin_changed(self):
        """Change vmin/vmax/under/over."""
        isvmin = self['isvmin'].isChecked()
        # Enable/Disable panels :
        self['vminW'].setEnabled(isvmin)
        # Vmin :
        self['isvmin'] = isvmin
        self['vmin'] = self['vmin'].value()
        self['under'] = color2json(self['under'])
        self.cbobjs['isvmin'] = self.cbviz.isvmin
        self.cbobjs['vmin'] = self.cbviz.vmin
        self.cbobjs['under'] = self.cbviz.under
        # Run object's update function :
        self.cbobjs.update()

    @_run_method_if_needed
    def _fcn_vmax_changed(self):
        """Change vmin/vmax/under/over."""
        isvmax = self['isvmax'].isChecked()
        self['vmaxW'].setEnabled(isvmax)
        # Vmax
        self['isvmax'] = isvmax
        self['vmax'] = self['vmax'].value()
        self['over'] = color2json(self['over'])
        self.cbobjs['isvmax'] = self.cbviz.isvmax
        self.cbobjs['vmax'] = self.cbviz.vmax
        self.cbobjs['over'] = self.cbviz.over
        # Run object's update function :
        self.cbobjs.update()

    @_run_method_if_needed
    def _fcn_limits(self):
        """Display/hide vmin/vmax."""
        self['limtxt'] = self['limTxt'].isChecked()
        self.cbobjs['limtxt'] = self.cbviz.limtxt

    # --------------------------------------------------------------------
    #                             TEXT
    # --------------------------------------------------------------------
    @_run_method_if_needed
    def _fcn_cb_title(self):
        """Change colorbar title."""
        self['cblabel'] = str(self['cblabel'].text())
        self.cbobjs['cblabel'] = self.cbviz.cblabel

    @_run_method_if_needed
    def _fcn_cb_txt_size(self):
        """Change colorbar text size."""
        self['cbtxtsz'] = self['cbTxtSz'].value()
        self.cbobjs['cbtxtsz'] = self.cbviz.cbtxtsz

    @_run_method_if_needed
    def _fcn_cb_txt_shift(self):
        """Change cblabel shift."""
        self['cbtxtsh'] = self['cbTxtSh'].value()
        self.cbobjs['cbtxtsh'] = self.cbviz.cbtxtsh

    @_run_method_if_needed
    def _fcn_txt_size(self):
        """Change text size for limits."""
        self['txtsz'] = self['txtSz'].value()
        self.cbobjs['txtsz'] = self.cbviz.txtsz

    @_run_method_if_needed
    def _fcn_txt_shift(self):
        """Change text shift."""
        self['txtsh'] = self['txtSh'].value()

    @_run_method_if_needed
    def _fcn_text_color(self):
        """Change text color."""
        txtcolor = color2json(self['txtCol'])
        self['txtcolor'] = txtcolor
        self.cbobjs['txtcolor'] = self.cbviz.txtcolor
        self.cbobjs['txtsh'] = self.cbviz.txtsh

    # --------------------------------------------------------------------
    #                             TEXT
    # --------------------------------------------------------------------
    def _fcn_cb_autoscale(self, *args, name=None):
        """Autoscale limits to data (Min, Max)."""
        # Select object if it's not the current one :
        if (name is not None) and self.cbobjs.name != name:
            self.select(name)
        # Disconnect clim buttons :
        self._trigger_cbar = False
        # Run the auto-scaling function :
        self['isvmin'] = False
        self['isvmax'] = False
        self.cbobjs['isvmin'] = False
        self.cbobjs['isvmax'] = False
        self.cbobjs.autoscale()
        # Set clim to the gui :
        self['climm'].setValue(self.cbobjs['clim'][0])
        self['climM'].setValue(self.cbobjs['clim'][1])
        # Check vmin/vmax values :
        self._check_vmin_vmax()
        # Set clim to the VisPy based colorbar :
        self['clim'] = self.cbobjs['clim']
        # Deactivate vmin/vmax :
        self['isvmin'].setChecked(False)
        self['isvmax'].setChecked(False)
        self['isvmin'] = self['isvmax'] = False
        self._trigger_cbar = True



"""Module reads and writes header and data for EDF data.

Poor man's version of
https://github.com/breuderink/eegtools/blob/master/eegtools/io/edfplus.py

Values are slightly different from those computed by FieldTrip, however they
are identical to those computed by Biosig and EDFBrowser. The difference is due
to the calibration.

"""
from logging import getLogger

from datetime import datetime
from math import floor
from re import findall
from numpy import empty, asarray, frombuffer, iinfo


lg = getLogger(__name__)
EDF_FORMAT = 'int16'  # by definition
edf_iinfo = iinfo(EDF_FORMAT)
DIGITAL_MAX = edf_iinfo.max
DIGITAL_MIN = -1 * edf_iinfo.max  # so that digital 0 = physical 0


def _assert_all_the_same(items):
    """Check that all the items in a list are the same."""
    assert all(items[0] == x for x in items)


class Edf:
    """Provide class EDF, which can be used to read the header and the data.

    Parameters
    ----------
    edffile : str
        Full path for the EDF file

    Attributes
    ----------
    h : dict
        disorganized header information

    """

    def __init__(self, edffile):
        """Init."""
        if isinstance(edffile, str):
            self.filename = edffile
            self._read_hdr()

    def _read_hdr(self):
        """Read header from EDF file.

        It only reads the header for internal purposes and adds a hdr.
        """
        with open(self.filename, 'rb') as f:

            hdr = {}
            assert f.tell() == 0
            assert f.read(8) == b'0       '

            # recording info
            hdr['subject_id'] = f.read(80).decode('utf-8').strip()
            hdr['recording_id'] = f.read(80).decode('utf-8').strip()

            # parse timestamp
            (day, month, year) = [int(x) for x in findall(
                r'(\d+)', f.read(8).decode('utf-8'))]
            (hour, minute, sec) = [int(x) for x in findall(
                r'(\d+)', f.read(8).decode('utf-8'))]
            hdr['start_time'] = datetime(year + 2000, month, day, hour, minute,
                                         sec)

            # misc
            hdr['header_n_bytes'] = int(f.read(8))
            f.seek(44, 1)  # reserved for EDF+
            hdr['n_records'] = int(f.read(8))
            hdr['record_length'] = float(f.read(8))  # in seconds
            nchannels = hdr['n_channels'] = int(f.read(4))

            # read channel info
            channels = range(hdr['n_channels'])
            hdr['label'] = [f.read(16).decode('utf-8').strip() for n in
                            channels]
            hdr['transducer'] = [f.read(80).decode('utf-8').strip()
                                 for n in channels]
            hdr['physical_dim'] = [f.read(8).decode('utf-8').strip() for n in
                                   channels]
            hdr['physical_min'] = asarray([float(f.read(8))
                                           for n in channels])
            hdr['physical_max'] = asarray([float(f.read(8))
                                           for n in channels])
            hdr['digital_min'] = asarray([float(f.read(8)) for n in channels])
            hdr['digital_max'] = asarray([float(f.read(8)) for n in channels])
            hdr['prefiltering'] = [f.read(80).decode('utf-8').strip()
                                   for n in channels]
            hdr['n_samples_per_record'] = [int(f.read(8)) for n in channels]
            f.seek(32 * nchannels, 1)  # reserved

            assert f.tell() == hdr['header_n_bytes']

            self.hdr = hdr

    def return_hdr(self):
        """Return the header for further use.

        Returns
        -------
        subj_id : str
            subject identification code
        start_time : datetime
            start time of the dataset
        s_freq : float
            sampling frequency
        chan_name : list of str
            list of all the channels
        n_samples : int
            number of samples in the dataset
        orig : dict
            additional information taken directly from the header
        """
        subj_id = self.hdr['subject_id']
        start_time = self.hdr['start_time']

        # Check that all channels have the same sampling rate
        # _assert_all_the_same(self.hdr['n_samples_per_record'])

        s_freq = (self.hdr['n_samples_per_record'][0] /
                  self.hdr['record_length'])

        chan_name = self.hdr['label']

        n_samples = (self.hdr['n_samples_per_record'][0] *
                     self.hdr['n_records'])

        return subj_id, start_time, s_freq, chan_name, n_samples, self.hdr

    def _read_dat(self, i_chan, begsam, endsam):
        """Read raw data from a single EDF channel.

        Reads only one channel at the time. Very initial implementation, very
        simple and probably not very fast

        Parameters
        ----------
        i_chan : int
            index of the channel to read
        begsam : int
            index of the first sample
        endsam : int
            index of the last sample

        Returns
        -------
        numpy.ndarray
            A vector with the data as written on file, in 16-bit precision
        """
        assert begsam < endsam

        begsam = float(begsam)
        endsam = float(endsam)

        n_sam_rec = self.hdr['n_samples_per_record']

        begrec = int(floor(begsam / n_sam_rec[i_chan]))
        begsam_rec = int(begsam % n_sam_rec[i_chan])

        endrec = int(floor(endsam / n_sam_rec[i_chan]))
        endsam_rec = int(endsam % n_sam_rec[i_chan])

        dat = empty(shape=(int(endsam) - int(begsam)), dtype='int16')
        i_dat = 0

        with open(self.filename, 'rb') as f:
            for rec in range(begrec, endrec + 1):
                if rec == begrec:
                    begpos_rec = begsam_rec
                else:
                    begpos_rec = 0

                if rec == endrec:
                    endpos_rec = endsam_rec
                else:
                    endpos_rec = n_sam_rec[i_chan]

                begpos = begpos_rec + sum(n_sam_rec) * rec + sum(
                    n_sam_rec[:i_chan])
                endpos = endpos_rec + sum(n_sam_rec) * rec + sum(
                    n_sam_rec[:i_chan])

                f.seek(begpos * 2 + self.hdr['header_n_bytes'])
                samples = f.read(2 * (endpos - begpos))

                i_dat_end = i_dat + endpos - begpos
                dat[i_dat:i_dat_end] = frombuffer(samples, dtype='<i2')
                i_dat = i_dat_end

        return dat

    def return_dat(self, chan, begsam, endsam):
        """Read data from an EDF file.

        Reads channel by channel, and adjusts the values by calibration.

        Parameters
        ----------
        chan : list of str
            index (indices) of the channels to read
        begsam : int
            index of the first sample
        endsam : int
            index of the last sample

        Returns
        -------
        numpy.ndarray
            A 2d matrix, where the first dimension is the channels and the
            second dimension are the samples.
        """
        hdr = self.hdr
        dig_min = hdr['digital_min']
        phys_min = hdr['physical_min']
        phys_range = hdr['physical_max'] - hdr['physical_min']
        dig_range = hdr['digital_max'] - hdr['digital_min']

        # assert all(phys_range > 0)
        # assert all(dig_range > 0)

        gain = phys_range / dig_range

        dat = empty(shape=(len(chan), endsam - begsam), dtype='float64')

        for i, i_chan in enumerate(chan):
            d = self._read_dat(i, begsam, endsam).astype('float64')
            dat[i, :] = (d - dig_min[i]) * gain[i] + phys_min[i]

        return dat

    def return_markers(self):
        """Return markers."""
        return []



"""write/Read hypnogram data.

Write hypnogram data
--------------------
-> write_hypno
    -> Export either using time code
        -> .txt, .csv, .xlsx
    -> Export using one sample per second
        -> .txt, .hyp

Read hypnogram data
-------------------
-> read_hypno
    -> detect_hypno_version
        -> Sample :
            -> .txt, .csv, .xlsx
        -> Tieme :
            -> .txt, .hyp
"""
import os
import logging
import numpy as np

from ..utils import vispy_array, transient
from ..io import is_pandas_installed, is_xlrd_installed

__all__ = ('oversample_hypno', 'write_hypno', 'read_hypno')

logger = logging.getLogger('visbrain')


###############################################################################
###############################################################################
#                              HYPNO CONVERSION
###############################################################################
###############################################################################

def hypno_time_to_sample(df, npts):
    """Convert the hypnogram from a defined timings to a number of samples.

    Parameters
    ----------
    df : pandas.DataFrame
        The data frame that contains timing.
    npts : int, array_like
        Number of time points in the final hypnogram. Alternatively, if npts is
        an array it will be interprated as the time vector.

    Returns
    -------
    hypno : array_like
        Hypnogram data of shape (npts,).
    time : array_like
        Time vector of shape (npts,).
    sf_hyp : float
        Sampling frequency of the hypnogram.
    """
    # Drop lines that contains * :
    drop_rows = np.char.find(np.array(df['Stage']).astype(str), '*')
    df = df.iloc[drop_rows.astype(bool)].copy()
    df.is_copy = False  # avoid pandas warning
    # Replace text by numerical values :
    to_replace = ['Wake', 'N1', 'N2', 'N3', 'REM', 'Art']
    values = [0, 1, 2, 3, 4, -1]
    df.replace(to_replace, values, inplace=True)
    # Get stages and time index :
    stages = np.array(df['Stage']).astype(str)
    time_idx = np.array(df['Time']).astype(float)
    # Compute time vector and sampling frequency :
    if isinstance(npts, np.ndarray):
        time = npts.copy()
    elif isinstance(npts, int):
        time = np.arange(npts) * time_idx[-1] / (npts - 1)
    sf_hyp = 1. / (time[1] - time[0])
    # Find closest time index :
    index = np.abs(time.reshape(-1, 1) - time_idx.reshape(1, -1))
    index = np.r_[0, index.argmin(0) + 1]
    # Fill the hypnogram :
    hypno = np.zeros((len(time),), dtype=int)
    for k in range(len(index) - 1):
        hypno[index[k]:index[k + 1]] = int(stages[k])
    return hypno, time, sf_hyp


def hypno_sample_to_time(hypno, time):
    """Convert the hypnogram from a number of samples to a defined timings.

    Parameters
    ----------
    hypno : array_like
        Hypnogram data.
    time : array_like
        The time vector.

    Returns
    -------
    df : pandas.DataFrame
        The data frame that contains all of the transient timings.
    """
    # Test if panda is installed :
    is_pandas_installed(True)
    import pandas as pd
    # Transient detection :
    _, tr, stages = transient(hypno, time)
    # Save the hypnogram :
    items = np.array(['Wake', 'N1', 'N2', 'N3', 'REM', 'Art'])
    return pd.DataFrame({'Stage': items[stages], 'Time': tr[:, 1]})


def oversample_hypno(hypno, n):
    """Oversample hypnogram.

    Parameters
    ----------
    hypno : array_like
        Hypnogram data of shape (N,) with N < n.
    n : int
        The destination length.

    Returns
    -------
    hypno : array_like
        The hypnogram of shape (n,)
    """
    # Get the repetition number :
    rep_nb = int(np.round(n / len(hypno)))

    # Repeat hypnogram :
    hypno = np.repeat(hypno, rep_nb)
    npts = len(hypno)

    # Check size
    if npts < n:
        hypno = np.append(hypno, hypno[-1] * np.ones((n - npts)))
    elif npts > n:
        hypno = hypno[0:n]

    return hypno.astype(int)


###############################################################################
###############################################################################
#                               WRITE HYPNO
###############################################################################
###############################################################################

def write_hypno(filename, hypno, version='time', sf=100., npts=1, window=1.,
                time=None, info=None):
    """Save hypnogram data.

    Parameters
    ----------
    filename : str
        Filename (with full path) of the file to save
    hypno : array_like
        Hypnogram array, same length as data
    sf : float | 100.
        Original sampling rate of the raw data
    npts : int | 1
        Original number of points in the raw data
    window : float | 1
        Time window (second) of each point in the hypno
        Default is one value per second
        (e.g. window = 30 = 1 value per 30 second)
    time : array_like | None
        The time vector.
    info : dict | None
        Additional informations to add to the file (prepend with *).
    """
    # Checking :
    assert isinstance(filename, str)
    assert isinstance(hypno, np.ndarray)
    assert version in ['time', 'sample']
    # Extract file extension :
    _, ext = os.path.splitext(filename)
    # Switch between time and sample version :
    if version == 'sample':  # v1 = sample
        # Take a down-sample version of the hypno :
        step = int(len(hypno) / np.round(npts / sf))
        hypno = hypno[::step].astype(int)
        # Export :
        if ext == '.txt':
            _write_hypno_txt_sample(filename, hypno, window=window)
        elif ext == '.hyp':
            _write_hypno_hyp_sample(filename, hypno, sf=sf, npts=npts)
    elif version == 'time':  # v2 = time
        # Get the DataFrame :
        df = hypno_sample_to_time(hypno, time)
        if isinstance(info, dict):
            is_pandas_installed(True)
            import pandas as pd
            info = {'*' + k: i for k, i in info.items()}
            df_info = pd.DataFrame({'Stage': list(info.keys()),
                                    'Time': list(info.values())})
            # df = df_info.append(df)
            df = pd.concat([df_info, df], ignore_index=True)
        if ext in ['.txt', '.csv']:
            df.to_csv(filename, header=None, index=None, sep='\t', mode='a')
        elif ext == '.xlsx':
            is_pandas_installed(True)
            is_xlrd_installed(True)
            import pandas as pd
            writer = pd.ExcelWriter(filename)
            df.to_excel(writer, sheet_name='Data', index=False, header=False)
            writer.save()
    logger.info("Hypnogram saved (%s)" % filename)


def _write_hypno_txt_sample(filename, hypno, window=1.):
    """Save hypnogram in txt file format (txt).

    Header is in file filename_description.txt

    Parameters
    ----------
    filename : str
        Filename (with full path) of the file to save
    hypno : array_like
        Hypnogram array, same length as data
    window : float | 1
        Time window (second) of each point in the hypno
        Default is one value per second
        (e.g. window = 30 = 1 value per 30 second)
    """
    base = os.path.basename(filename)
    dirname = os.path.dirname(filename)
    descript = os.path.join(dirname, os.path.splitext(
        base)[0] + '_description.txt')

    # Save hypno
    np.savetxt(filename, hypno, fmt='%s')

    # Save header file
    hdr = np.array([['time ' + str(window)], ['W 0'], ['N1 1'], ['N2 2'],
                    ['N3 3'], ['REM 4'], ['Art -1']]).flatten()
    np.savetxt(descript, hdr, fmt='%s')


def _write_hypno_hyp_sample(filename, hypno, sf=100., npts=1):
    """Save hypnogram in Elan file format (hyp).

    Parameters
    ----------
    filename : str
        Filename (with full path) of the file to save
    hypno : array_like
        Hypnogram array, same length as data
    sf : float | 100.
        Original sampling rate of the raw data
    npts : int | 1
        Original number of points in the raw data
    """
    hypno[hypno == 4] = 5

    hdr = np.array([['time_base 1.000000'],
                    ['sampling_period ' + str(np.round(1 / sf, 8))],
                    ['epoch_nb ' + str(int(npts / sf))],
                    ['epoch_list']]).flatten()

    # Save
    export = np.append(hdr, hypno.astype(str))
    np.savetxt(filename, export, fmt='%s')


###############################################################################
###############################################################################
#                                 READ HYPNO
###############################################################################
###############################################################################


def read_hypno(filename, time=None, datafile=None):
    """Load hypnogram file.

    Sleep stages in the hypnogram should be scored as follow
    see Iber et al. 2007

    Wake:   0
    N1:     1
    N2:     2
    N3:     3
    REM:    4
    Art:    -1  (optional)

    Parameters
    ----------
    filename : string
        Filename (with full path) to hypnogram file.
    time : array_like | None
        The time vector (used to interpolate Excel files).
    datafile : string | None
        Filename (with full path) to the data file.

    Returns
    -------
    hypno : array_like
        The hypnogram vector in its original length.
    sf_hyp: float
        The hypnogram original sampling frequency (Hz)
    """
    # Test if file exist :
    assert os.path.isfile(filename)

    # Extract file extension :
    file, ext = os.path.splitext(filename)

    # Load the hypnogram :
    if ext == '.hyp':  # v1 = ELAN
        hypno, sf_hyp = _read_hypno_hyp_sample(filename)
    elif ext == '.edf':  # v1 = EDF+
        hypno, sf_hyp = _read_hypno_edf_sample(filename, datafile)
    elif ext in ['.txt', '.csv']:  # [v1, v2] = TXT / CSV
        header = os.path.splitext(filename)[0] + '_description.txt'
        if os.path.isfile(header):  # if there's a header -> v1
            hypno, sf_hyp = _read_hypno_txt_sample(filename)
        else:  # v2
            import pandas as pd
            df = pd.read_csv(filename, delim_whitespace=True, header=None,
                             names=['Stage', 'Time'])
            hypno, _, sf_hyp = hypno_time_to_sample(df, len(time))
    elif ext == '.xlsx':  # v2 = Excel
        import pandas as pd
        df = pd.read_excel(filename, header=None, names=['Stage', 'Time'])
        hypno, _, sf_hyp = hypno_time_to_sample(df, len(time))

    logger.info("Hypnogram successfully loaded (%s)" % filename)

    return vispy_array(hypno), sf_hyp


def _read_hypno_hyp_sample(path):
    """Read Elan hypnogram (hyp).

    Parameters
    ----------
    path : str
        Filename(with full path) to Elan .hyp file

    Returns
    -------
    hypno : array_like
        The hypnogram vector in its original length.
    sf_hyp : float
        The hypnogram original sampling frequency (Hz)
    """
    hyp = np.genfromtxt(path, delimiter='\n', usecols=[0],
                        dtype=None, skip_header=0, encoding='utf-8')

    # Get sampling frequency of hypnogram
    sf_hyp = 1 / float(hyp[0].split()[1])

    # Extract hypnogram values
    # TODO: changed from np.int to int
    hypno = np.array(hyp[4:], dtype=int)

    # Replace values according to Iber et al 2007
    hypno[hypno == -2] = -1
    hypno[hypno == 4] = 3
    hypno[hypno == 5] = 4

    return hypno, sf_hyp


def _read_hypno_txt_sample(path):
    """Read text files (.txt / .csv) hypnogram.

    Parameters
    ----------
    path : str
        Filename(with full path) to hypnogram (.txt)

    Returns
    -------
    hypno : array_like
        The hypnogram vector in its original length.
    sf_hyp : float
        The hypnogram original sampling frequency (Hz)
    """
    assert os.path.isfile(path)

    file, ext = os.path.splitext(path)

    header = file + '_description.txt'
    assert os.path.isfile(header)

    # Load header file
    labels = np.genfromtxt(header, dtype=str, delimiter=" ", usecols=0,
                           encoding='utf-8')
    values = np.genfromtxt(header, dtype=float, delimiter=" ", usecols=1,
                           encoding='utf-8')
    desc = {label: row for label, row in zip(labels, values)}

    # Get sampling frequency of hypnogram
    sf_hyp = 1. / float(desc['time'])

    # Load hypnogram file
    hyp = np.genfromtxt(path, delimiter='\n', usecols=[0],
                        dtype=None, skip_header=0, encoding='utf-8')

    if not np.issubdtype(hyp.dtype, np.integer):
        hypno = np.array([s for s in hyp if s.lstrip('-').isdigit()],
                         dtype=int)
    else:
        hypno = hyp.astype(int)

    hypno = swap_hyp_values(hypno, desc)

    return hypno, sf_hyp


def _read_hypno_edf_sample(hypno_file_path, data_file_path):
    """Read hypnogram files which are formatted according to EDF+.

    See file specifications (see https://www.edfplus.info/specs/index.html).
    The function was specifically developed to read hypnograms which are part
    of the Physionet Sleep Database (https://physionet.org/pn4/sleep-edfx/).
    The Physionet Sleep Database contains 61 polysomnograms (PSGs) with
    accompanying hypnograms. The Physionet polysomnogram files are using edf
    format, while the hypnograms use EDF+. When using EDF+ for hypnograms, the
    data records contain Timestamped Annotation Lists (TALs). Each TAL consists
    of an onset, a duration, and a sleep stage. The following assumptions have
    been made:
    - the scoring period for the hypnogram is equal to the number of seconds
      per data record as given by header information of the accompanying
      polysomnogram file
    - all EDF+ hypnogram files use the same scoring for sleep stages which are
      in the function converted to the values used by Visbrain:
            EDF+ score                    Visbrain
          Sleep stage ?                      -1
          Movement time                      -1
          Sleep stage W                       0
          Sleep stage 1                       1
          Sleep stage 2                       2
          Sleep stage 3                       3
          Sleep stage 4                       3
          Sleep stage R                       4

    The Physionet hypnogram files (EDF+) sometimes cover a slightly longer
    total period than the polysomnograph files, where the score for the last
    part is set to "Sleep stage ?". This function will only read the hypnogram
    files so that only the time period of the polysomnograph is covered.

    The function will check whether the hypnogram file is the appropriate
    accompanying file to the polysomnogranph file by comparing the "subjcet_id"
    as well as the "recording_id" of the two files.

    Parameters
    ----------
    data_file_path : str
        Filename(with full path) to data(.edf)
    hypno_file_path : str
        Filename(with full path) to corresponding hypnogram(.edf)

    Returns
    -------
    hypno : array_like
        The hypnogram vector in its original length.
    sf_hyp : float
        The hypnogram original sampling frequency (Hz)
    """
    if not isinstance(data_file_path, str):
        raise IOError("Reading EDF+ need the path to the data file.")
    data_file_path, _ = os.path.splitext(data_file_path)

    with open(data_file_path + '.edf', 'rb') as f:  # open edf data file
        hdr1 = {}
        assert f.tell() == 0
        assert f.read(8) == b'0       '

        # Recording info (patient info and date and time)
        hdr1['subject_id'] = f.read(80).decode('utf-8').strip()
        hdr1['recording_id'] = f.read(80).decode('utf-8').strip()

        f.seek(68, 1)
        hdr1['n_records'] = int(f.read(8))
        hdr1['record_length'] = float(f.read(8))  # in seconds
        end_file = str(int(hdr1['n_records'] * hdr1['record_length']))

    with open(hypno_file_path, 'rb') as f:  # open edf hypnogram file
        hdr2 = {}
        assert f.tell() == 0
        assert f.read(8) == b'0       '

        # Recording info (patient info and date and time)
        hdr2['subject_id'] = f.read(80).decode('utf-8').strip()
        hdr2['recording_id'] = f.read(80).decode('utf-8').strip()

        # Compare the patient info and recording date of the two files
        if not (hdr1['subject_id'] == hdr2['subject_id'] and hdr1[
                'recording_id'] == hdr2['recording_id']):
            raise IOError("Hypnogram file does not match polysomnograph file")

        # skip records not required
        f.seek(16, 1)
        # read bytes in header of hypnogram file
        hdr2['header_n_bytes'] = int(f.read(8))

        # go to the end of the header info```
        f.seek(hdr2['header_n_bytes'])

        data_hypno = f.read().decode('utf-8')  # read the data

    # number of secs per data record used for score period
    time = hdr1['record_length']
    data_hypno_spl = data_hypno.split('\x00')
    ln = len(data_hypno_spl)
    tr = {'Sleep stage ?': -1, 'Movement time': -1, 'Sleep stage W': 0,
          'Sleep stage 1': 1, 'Sleep stage 2': 2, 'Sleep stage 3': 3,
          'Sleep stage 4': 3, 'Sleep stage R': 4}
    hypno_s = []
    for i in range(ln):
        in_start = data_hypno_spl[i].find('\x15')
        if in_start == -1:
            continue
        else:
            in_stop = data_hypno_spl[i].find('\x14')
            onset = data_hypno_spl[i][1:in_start]
            duration = data_hypno_spl[i][in_start + 1:in_stop]

            if int(onset) + int(duration) >= int(end_file):
                duration = str(int(end_file) - int(onset))

            sleepstage = data_hypno_spl[i][in_stop + 1:-1]

            nr = int(int(duration) / 30)
            entry = [tr[sleepstage]] * nr
            hypno_s.extend(entry)

    hypno_s = np.array(hypno_s).astype(int)
    sf_hyp = 1. / time
    return hypno_s, sf_hyp


def swap_hyp_values(hypno, desc):
    """Swap values in hypnogram vector.

    Sleep stages in the hypnogram should be scored as follow
    see Iber et al. 2007

    e.g from the DREAM bank EDF database
    Stage   Orig. val    New val
    W       5           0
    N1      3           1
    N2      2           2
    N3      1           3
    REM     0           4

    Parameters
    ----------
    hypno : array_like
        The hypnogram vector
    description : str
        Path to a .txt file containing labels and values of each sleep
        stage separated by a space

    Returns
    -------
    hypno_s : array_like
        Hypnogram with swapped values
    """
    # Swap values
    hypno_s = -1 * np.ones(shape=(hypno.shape), dtype=int)

    if 'Art' in desc:
        hypno_s[hypno == desc['Art']] = -1
    if 'Nde' in desc:
        hypno_s[hypno == desc['Nde']] = -1
    if 'Mt' in desc:
        hypno_s[hypno == desc['Mt']] = -1
    if 'W' in desc:
        hypno_s[hypno == desc['W']] = 0
    if 'N1' in desc:
        hypno_s[hypno == desc['N1']] = 1
    if 'N2' in desc:
        hypno_s[hypno == desc['N2']] = 2
    if 'N3' in desc:
        hypno_s[hypno == desc['N3']] = 3
    if 'N4' in desc:
        hypno_s[hypno == desc['N4']] = 3
    if 'REM' in desc:
        hypno_s[hypno == desc['REM']] = 4

    return hypno_s



"""Main class for sleep menus managment."""
import os

import numpy as np
from PyQt5 import QtWidgets

from sleep_scoring.utils import HelpMenu
from sleep_scoring.io import (dialog_save, dialog_load, write_fig_hyp, write_csv,
                         write_txt, write_hypno, read_hypno,
                         annotations_to_array, oversample_hypno,
                         save_config_json)


class UiMenu(HelpMenu):
    """Main class for sleep menus managment."""

    def __init__(self):
        """Init."""
        base = 'http://visbrain.org/sleep.html'
        sections = {'Sleep': base,
                    'Time-frequency': base + '#time-frequency',
                    'Hypnogram scoring': base + '#hypnogram-scoring',
                    'Detections': base + '#apply-semi-automatic-detection',
                    'Annotations': base + '#import-add-and-save-annotations'}
        HelpMenu.__init__(self, sections)
        # _____________________________________________________________________
        #                                 SAVE
        # _____________________________________________________________________
        # Hypnogram :
        self.menuSaveHypnogramData.triggered.connect(self.saveHypData)
        self.menuSaveHypnogramFigure.triggered.connect(self._save_hyp_fig)
        # Stats info table :
        self.menuSaveInfoTable.triggered.connect(self._save_info_table)
        # Scoring table :
        self.menuSaveScoringTable.triggered.connect(self._save_scoring_table)
        # Detections :
        self.menuSaveDetectAll.triggered.connect(self._save_all_detect)
        self.menuSaveDetectSelected.triggered.connect(self._save_select_detect)
        # Sleep GUI config :
        self.menuSaveConfig.triggered.connect(self._save_config)
        # Annotations :
        self.menuSaveAnnotations.triggered.connect(self._save_annotation_table)
        # Screenshot :
        self.menuSaveScreenshot.triggered.connect(self._fcn_gui_screenshot)

        # _____________________________________________________________________
        #                                 LOAD
        # _____________________________________________________________________
        # Load hypnogram :
        self.menuLoadHypno.triggered.connect(self._load_hypno)
        # Sleep GUI config :
        self.menuLoadConfig.triggered.connect(self._load_config)
        # Detections :
        self.menuLoadDetectAll.triggered.connect(self._load_detect_all)
        self.menuLoadDetectSelect.triggered.connect(self._load_detect_select)
        # Annotations :
        self.menuLoadAnnotations.triggered.connect(self._load_annotation_table)

        # _____________________________________________________________________
        #                                 EXIT
        # _____________________________________________________________________
        self.menuExit.triggered.connect(QtWidgets.qApp.quit)

        # _____________________________________________________________________
        #                              DISPLAY
        # _____________________________________________________________________
        # Quick settings panel :
        self.menuDispSettings.triggered.connect(self._disptog_settings)
        self.q_widget.setVisible(True)
        # Spectrogram :
        self.menuDispSpec.triggered.connect(self._disptog_spec)
        # Hypnogram :
        self.menuDispHypno.triggered.connect(self._disptog_hyp)
        # Time axis :
        self.menuDispTimeax.triggered.connect(self._disptog_timeax)
        # Navigation bar :
        self.menuDispNavbar.triggered.connect(self._disptog_navbar)
        # Time indicators :
        self.menuDispIndic.triggered.connect(self._disptog_indic)
        # Topoplot :
        self.menuDispTopo.triggered.connect(self._disptog_topo)
        # Zoom :
        self.menuDispZoom.triggered.connect(self._disptog_zoom)

    ###########################################################################
    ###########################################################################
    #                           SAVE
    ###########################################################################
    ###########################################################################

    # ______________________ HYPNOGRAM ______________________
    def saveHypData(self, *args, filename=None, reply=None):  # noqa
        def prompt_write_hypno(dialog_ext, version, filename, format):
            # Open dialog window :
            if filename is None:
                filename = dialog_save(self, 'Save File', hyp_file, dialog_ext +
                                    ";;All files (*.*)")
            if filename:
                info = {'Duration_sec': self._N * 1 / self._sfori}
                if isinstance(self._file, str):
                    info['Datafile'] = self._file

                filename_no_ext, ext = os.path.splitext(filename)
                write_hypno(filename_no_ext + format + ext, self._hypno, version=version, sf=self._sfori,
                            npts=self._N, time=self._time, info=info)
        
        """Save the hypnogram data either in a hyp or txt file."""
        # Define default filename for the hypnogram :
        if not isinstance(self._file, str):
            hyp_file = 'hypno'
        else:
            hyp_file = os.path.basename(self._file) + '_hypno'
        # Version switch :
        if reply is None:
            msg = ("Since release 0.4 of Sleep, hypnograms are exported using stage "
                   "durations rather than point-per-second. This new format "
                   "avoids potential errors caused by downsampling and "
                   "confusion in the values assigned to each sleep stage. \n\n"
                   "Click 'Yes' to use the new (v2) format and 'No' to use the old (v1) "
                   "format. Alternatively, you can save both formats at once, and "
                   "Sleep will label them appropriately.")
            
            msg_box = QtWidgets.QMessageBox()
            button_yes = QtWidgets.QMessageBox.Yes
            button_no = QtWidgets.QMessageBox.No
            button_both = QtWidgets.QPushButton("Save both formats")
            msg_box.addButton(button_yes)
            msg_box.addButton(button_no)
            msg_box.addButton(button_both, QtWidgets.QMessageBox.ActionRole)
            msg_box.setText(msg)

            reply = msg_box.exec()

        if reply == QtWidgets.QMessageBox.No:  # v1 = sample
            dialog_ext = "Text file (*.txt);;Elan file (*.hyp)"
            version = 'sample'
            prompt_write_hypno(dialog_ext=dialog_ext, version=version, filename=filename, format="")
        elif reply == QtWidgets.QMessageBox.Yes:  # v2 = time
            dialog_ext = ("Text file (*.txt);;Csv file (*.csv);;Excel file "
                          "(*.xlsx)")
            version = 'time'
            prompt_write_hypno(dialog_ext=dialog_ext, version=version, filename=filename, format="")
        else:
            # v1
            dialog_ext = "Text file (*.txt);;Elan file (*.hyp)"
            version = 'sample'
            prompt_write_hypno(dialog_ext=dialog_ext, version=version, filename=filename, format="-v1")
            # v2
            dialog_ext = ("Text file (*.txt);;Csv file (*.csv);;Excel file "
                          "(*.xlsx)")
            version = 'time'
            prompt_write_hypno(dialog_ext=dialog_ext, version=version, filename=filename, format="-v2")
        

    def _save_hyp_fig(self, *args, filename=None, **kwargs):
        """Save a 600 dpi .png figure of the hypnogram."""
        if filename is None:
            filename = dialog_save(self, 'Save Hypnogram figure', 'hypno',
                                   "PNG (*.png);;All files (*.*)")
        if filename:
            hypno = self._hypno.copy()
            grid = self._slGrid.isChecked()
            ascolor = self._PanHypnoColor.isChecked()
            write_fig_hyp(hypno, self._sf, file=filename,
                          start_s=self._toffset, grid=grid, ascolor=ascolor)

    # ______________________ STATS INFO TABLE ______________________
    def _save_info_table(self, *args, filename=None):
        """Export stat info."""
        # Get file name :
        if filename is None:
            filename = dialog_save(self, 'Save file', 'statsinfo',
                                   "CSV file (*.csv);;Text file (*.txt);;"
                                   "All files (*.*)")
        if filename:
            file, ext = os.path.splitext(filename)
            if ext.find('csv') + 1:
                write_csv(file + '.csv', zip(self._keysInfo, self._valInfo))
            elif ext.find('txt') + 1:
                write_txt(file + '.txt', zip(self._keysInfo, self._valInfo))

    # ______________________ SCORING TABLE ______________________
    def _save_scoring_table(self, *args, filename=None):
        """Export score info."""
        # Read Table
        row_count = self._scoreTable.rowCount()
        sta_ind, end_ind, stage = [], [], []
        for row in np.arange(row_count):
            sta_ind.append(str(self._scoreTable.item(row, 0).text()))
            end_ind.append(str(self._scoreTable.item(row, 1).text()))
            stage.append(str(self._scoreTable.item(row, 2).text()))
        # Get file name :
        if filename is None:
            filename = dialog_save(self, 'Save file', 'scoring_info',
                                   "CSV file (*.csv);;Text file (*.txt);;"
                                   "All files (*.*)")
        if filename:
            file, ext = os.path.splitext(filename)
            if ext.find('csv') + 1:
                write_csv(file + '.csv', zip(sta_ind, end_ind, stage))
            elif ext.find('txt') + 1:
                write_txt(file + '.txt', zip(sta_ind, end_ind, stage))

    # ______________________ DETECTIONS ______________________
    def _check_detect_menu(self):
        """Activate/Deactivate the saving detection menu."""
        self.menuSaveDetections.setEnabled(bool(self._detect))

    def _save_all_detect(self, *args, filename=None):
        """Export all detections."""
        # Get file name :
        if filename is None:
            filename = dialog_save(self, 'Save all detections', 'detections',
                                   "NumPy (*.npy);;All files (*.*)")
        if filename:
            file = os.path.splitext(str(filename))[0]
            np.save(file + '.npy', self._detect.dict)

    def _save_select_detect(self, *args, filename=None):
        """Export selected detection."""
        channel, method = self._get_current_chan_type()
        # Read Table
        row_count = self._DetectLocations.rowCount()
        sta_ind = [channel, '', 'Time index (s)']
        end_ind = [method, '', 'Time index (s)']
        duration = ['', '', 'Duration (s)']
        stage = ['', '', 'Sleep stage']
        for row in np.arange(row_count):
            sta_ind.append(str(self._DetectLocations.item(row, 0).text()))
            end_ind.append(str(self._DetectLocations.item(row, 1).text()))
            duration.append(str(self._DetectLocations.item(row, 2).text()))
            stage.append(str(self._DetectLocations.item(row, 3).text()))
        # Get file name :
        saveas = "locinfo" + '_' + channel + '-' + method
        if filename is None:
            filename = dialog_save(self, 'Save ' + method + ' detection',
                                   saveas, "CSV file (*.csv);;Text file"
                                   " (*.txt);;All files (*.*)")
        if filename:
            file, ext = os.path.splitext(filename)
            file += '_' + channel + '-' + method
            zp = zip(sta_ind, end_ind, duration, stage)
            if ext.find('csv') + 1:
                write_csv(file + '.csv', zp)
            elif ext.find('txt') + 1:
                write_txt(file + '.txt', zp)

    # ______________________ SLEEP GUI CONFIG ______________________
    def _save_config(self, *args, filename=None):
        """Save a config file (*.txt) containing several display parameters."""
        if filename is None:
            filename = dialog_save(self, 'Save config File', 'config',
                                   "JSON file (*.json);;Text file (*.txt)")
        if filename:
            config = {}
            # Get channels visibility / amplitude :
            viz, amp = [], []
            for i, k in enumerate(self._chanChecks):
                viz.append(k.isChecked())
                amp.append(self._ymaxSpin[i].value())
            config['Channel_Names'] = self._channels
            config['Channel_Visible'] = viz
            config['Channel_Amplitude'] = amp
            # config['AllAmpMin'] = self._PanAllAmpMin.value()
            # config['AllAmpMax'] = self._PanAllAmpMax.value()
            config['SymAmp'] = self._PanAmpSym.isChecked()
            config['AutoAmp'] = self._PanAmpAuto.isChecked()
            # Spectrogram :
            config['Spec_Visible'] = self.menuDispSpec.isChecked()
            config['Spec_Method'] = self._PanSpecMethod.currentIndex()
            config['Spec_Length'] = self._PanSpecNfft.value()
            config['Spec_Overlap'] = self._PanSpecStep.value()
            config['Spec_Cmap'] = self._PanSpecCmap.currentIndex()
            config['Spec_CmapInv'] = self._PanSpecCmapInv.isChecked()
            config['Spec_Chan'] = self._PanSpecChan.currentIndex()
            config['Spec_Fstart'] = self._PanSpecFstart.value()
            config['Spec_Fend'] = self._PanSpecFend.value()
            config['Spec_Con'] = self._PanSpecCon.value()
            config['Spec_Interp'] = self._PanSpecInterp.currentIndex()
            # Hypnogram/time axis/navigation/topo/indic/zoom :
            config['Hyp_Visible'] = self.menuDispHypno.isChecked()
            config['Time_Visible'] = self.menuDispTimeax.isChecked()
            config['Topo_Visible'] = self.menuDispTopo.isChecked()
            config['Nav_Visible'] = self.menuDispNavbar.isChecked()
            config['Indic_Visible'] = self.menuDispIndic.isChecked()
            config['Zoom_Visible'] = self.menuDispZoom.isChecked()
            config['Hyp_Lw'] = self._PanHypnoLw.value()
            config['Hyp_Color'] = self._PanHypnoColor.isChecked()
            # Navigation bar properties :
            config['Slider'] = self._SlVal.value()
            config['Step'] = self._SigSlStep.value()
            config['Window'] = self._SigWin.value()
            config['Goto'] = self._SlGoto.value()
            config['Magnify'] = self._slMagnify.isChecked()
            config['AbsTime'] = self._slAbsTime.isChecked()
            config['Grid'] = self._slGrid.isChecked()
            config['Unit'] = self._slRules.currentIndex()
            save_config_json(filename, config)

    # ______________________ ANNOTATION TABLE ______________________
    def _save_annotation_table(self, *args, filename=None):
        """Export annotation table."""
        # Read Table
        row_count = self._AnnotateTable.rowCount()
        sta_ind, end_ind, annot = [], [], []
        for row in np.arange(row_count):
            sta_ind.append(str(self._AnnotateTable.item(row, 0).text()))
            end_ind.append(str(self._AnnotateTable.item(row, 1).text()))
            annot.append(str(self._AnnotateTable.item(row, 2).text()))
        # Get file name :
        if filename is None:
            filename = dialog_save(self, 'Save annotations', 'annotations',
                                   "CSV file (*.csv);;Text file (*.txt);;"
                                   "All files (*.*)")
        if filename:
            file, ext = os.path.splitext(filename)
            if ext.find('csv') + 1:
                write_csv(file + '.csv', zip(sta_ind, end_ind, annot))
            elif ext.find('txt') + 1:
                write_txt(file + '.txt', zip(sta_ind, end_ind, annot))

    # ______________________ SCREENSHOT ______________________
    def _fcn_gui_screenshot(self):
        """Screenshot using the GUI."""
        self.show_gui_screenshot()

    ###########################################################################
    ###########################################################################
    #                                    LOAD
    ###########################################################################
    ###########################################################################

    def _load_hypno(self, *args, filename=None):
        """Load a hypnogram."""
        # Define default filename for the hypnogram :
        if not isinstance(self._file, str):
            hyp_file = 'hypno'
        else:
            hyp_file = os.path.basename(self._file) + '_hypno'
        # Get filename :
        if filename is None:
            filename = dialog_load(self, 'Load hypnogram File', hyp_file,
                                   "Text file (*.txt);;CSV file (*.csv);;"
                                   "Elan file (*.hyp);;Excel file (*.xlsx);;"
                                   "All files (*.*)")
        if filename:
            # Load the hypnogram :
            self._hypno, _ = read_hypno(filename, time=self._time)
            self._hypno = oversample_hypno(self._hypno, self._N)[::self._dsf]
            self._hyp.set_data(self._sf, self._hypno, self._time)
            # Update info table :
            self._fcn_info_update()
            # Update scoring table :
            self._fcn_hypno_to_score()
            self._fcn_score_to_hypno()

    def _load_config(self, *args, filename=None):
        """Load a config file (*.txt) containing several display parameters."""
        import json
        if not filename:
            filename = dialog_load(self, 'Load config File', 'config',
                                   "Text file (*.txt);;All files (*.*)")
        if filename:
            with open(filename) as f:
                # Load the configuration file :
                config = json.load(f)

                def _try(string, self=self, config=config):
                    """Execute the string.

                    This function insure backward compatibility for loading the
                    configuration file.
                    """
                    try:
                        exec(string)
                    except:
                        pass

                # Channels
                for i, k in enumerate(self._chanChecks):
                    self._chanChecks[i].setChecked(
                        config['Channel_Visible'][i])
                    self._ymaxSpin[i].setValue(config['Channel_Amplitude'][i])
                # Amplitudes :
                # _try("self._PanAllAmpMin.setValue(config['AllAmpMin'])")
                # _try("self._PanAllAmpMax.setValue(config['AllAmpMax'])")
                _try("self._PanAmpSym.setChecked(config['SymAmp'])")
                _try("self._PanAmpAuto.setChecked(config['AutoAmp'])")
                # Spectrogram
                _try("self.menuDispSpec.setChecked(config['Spec_Visible'])")
                _try("self._PanSpecMethod.setCurrentIndex("
                     "config['Spec_Method'])")
                _try("self._PanSpecNfft.setValue(config['Spec_Length'])")
                _try("self._PanSpecStep.setValue(config['Spec_Overlap'])")
                _try("self._PanSpecCmap.setCurrentIndex(config['Spec_Cmap'])")
                _try("self._PanSpecCmapInv.setChecked(config['Spec_CmapInv'])")
                _try("self._PanSpecChan.setCurrentIndex(config['Spec_Chan'])")
                _try("self._PanSpecFstart.setValue(config['Spec_Fstart'])")
                _try("self._PanSpecFend.setValue(config['Spec_Fend'])")
                _try("self._PanSpecCon.setValue(config['Spec_Con'])")
                _try("self._PanSpecInterp.setCurrentIndex("
                     "config['Spec_Interp'])")
                # Hypnogram/time axis/navigation/topo/indic/zoom :
                _try("self.menuDispHypno.setChecked(config['Hyp_Visible'])")
                _try("self.menuDispTimeax.setChecked(config['Time_Visible'])")
                _try("self.menuDispTopo.setChecked(config['Topo_Visible'])")
                _try("self.menuDispNavbar.setChecked(config['Nav_Visible'])")
                _try("self.menuDispIndic.setChecked(config['Indic_Visible'])")
                _try("self._PanHypnoLw.setValue(config['Hyp_Lw'])")
                _try("self._PanHypnoColor.setChecked(config['Hyp_Color'])")
                # Navigation bar properties :
                _try("self._SlVal.setValue(config['Slider'])")
                _try("self._SigSlStep.setValue(config['Step'])")
                _try("self._SigWin.setValue(config['Window'])")
                _try("self._SlGoto.setValue(config['Goto'])")
                _try("self._slMagnify.setChecked(config['Magnify'])")
                _try("self._slAbsTime.setChecked(config['AbsTime'])")
                _try("self._slGrid.setChecked(config['Grid'])")
                _try("self._slRules.setCurrentIndex(config['Unit'])")
                # Update display
                self._fcn_chan_viz()
                self._fcn_chan_amplitude()
                self._fcn_spec_set_data()
                self._disptog_spec()
                self._disptog_hyp()
                self._disptog_timeax()
                self._disptog_topo()
                self._disptog_indic()
                self._disptog_zoom()
                self._fcn_grid_toggle()
                self._fcn_update_amp_info()
                self._fcn_chan_auto_amp()
                self._fcn_chan_sym_amp()
                self._fcn_set_hypno_lw()
                self._fcn_set_hypno_color()

    def _load_detect_all(self, *args, filename=None):
        """Load all detections."""
        # Dialog window for detection file :
        if filename is None:
            filename = dialog_load(self, "Import detections", '',
                                   "NumPy (*.npy);;All files (*.*)")
        self._detect.dict = np.ndarray.tolist(np.load(filename))
        # Made canvas visbles :
        for k in self._detect:
            if self._detect[k]['index'].size:
                # Get channel number :
                idx = self._channels.index(k[0])
                self._canvas_set_visible(idx, True)
                self._chan.visible[idx] = True
        # Plot update :
        self._fcn_slider_move()
        self._loc_line_report()
        self._check_detect_menu()

    def _load_detect_select(self, *args, filename=None):
        """Load a specific detection."""
        # Get file name :
        if filename is None:
            filename = dialog_load(self, "Import table", '',
                                   "CSV file (*.csv);;Text file (*.txt);;"
                                   "All files (*.*)")
        if filename:
            # Get channel / method from file name :
            (chan, meth) = filename.split('_')[-1].split('.')[0].split('-')
            # Load the file :
            (st, end) = np.genfromtxt(filename, delimiter=',',
                                      encoding='utf-8')[3::, 0:2].T
            # Sort by starting index :
            idxsort = np.argsort(st)
            st, end = st[idxsort], end[idxsort]
            # Concatenate (starting, ending) index :
            index = np.c_[st, end]
            # Convert into index :
            index = np.round(index * self._sf).astype(int)
            # Set index :
            self._detect[(chan, meth)]['index'] = index
            # Plot update :
            self._fcn_slider_move()
            self._loc_line_report()
            self._check_detect_menu()

    def _load_annotation_table(self, *args, filename=None):
        """Load annotations."""
        # Get file name :
        if filename is None:
            filename = dialog_load(self, "Import annotations", '',
                                   "CSV file (*.csv);;Text file (*.txt);;"
                                   "All files (*.*)")
        # Clean annotations :
        self._AnnotateTable.setRowCount(0)
        start, end, annot = annotations_to_array(filename)

        # Fill table :
        self._AnnotateTable.setRowCount(len(start))
        # File the table :
        for k, (s, e, a) in enumerate(zip(start, end, annot)):
            # Starting index :
            self._AnnotateTable.setItem(
                k, 0, QtWidgets.QTableWidgetItem(str(s)))
            # Ending index :
            self._AnnotateTable.setItem(
                k, 1, QtWidgets.QTableWidgetItem(str(e)))
            # Text :
            self._AnnotateTable.setItem(
                k, 2, QtWidgets.QTableWidgetItem(str(a)))
            # Set the current tab to the annotation tab :
            self.QuickSettings.setCurrentIndex(5)
        # Set markers :
        middle = (start.astype(np.float32) + end.astype(np.float32)) / 2
        self._annot_mark = middle
        self._fcn_slider_move()

    ###########################################################
    ###########################################################
    #                            DISPLAY
    ###########################################################
    ###########################################################

    def _disptog_settings(self):
        """Toggle method for display / hide the settings panel.

        Shortcut : CTRL + D
        """
        viz = self.q_widget.isVisible()
        self.q_widget.setVisible(not viz)
        # Set topo widget larger if settings panel hide :
        self._topoW.setMaximumWidth(300 * (1 + .7 * viz))

    def _disptog_spec(self):
        """Toggle method for display / hide the spectrogram.

        Shortcut : S
        """
        viz = self.menuDispSpec.isChecked()
        self._SpecW.setVisible(viz)
        self._specLabel.setVisible(viz)

    def _disptog_hyp(self):
        """Toggle method for display / hide the hypnogram.

        Shortcut : H
        """
        viz = self.menuDispHypno.isChecked()
        self._HypW.setVisible(viz)
        self._hypLabel.setVisible(viz)

    def _disptog_navbar(self):
        """Toggle method for display / hide the navigation bar.

        Shortcut : P
        """
        self._slFrame.hide() if self._slFrame.isVisible(
        ) else self._slFrame.show()

    def _disptog_timeax(self):
        """Toggle method for display / hide the time axis.

        Shortcut : X
        """
        viz = self.menuDispTimeax.isChecked()
        self._TimeAxisW.setVisible(viz)
        self._timeLabel.setVisible(viz)

    def _disptog_topo(self):
        """Toggle method for display / hide the topoplot.

        Shortcut : T
        """
        viz = self.menuDispTopo.isChecked()
        self._topoW.setVisible(viz)
        self._PanTopoVizW.setEnabled(viz)
        if viz:
            self._fcn_topo_settings()
            self._fcn_slider_move()

    def _disptog_indic(self):
        """Toggle method for display / hide the time indicators."""
        viz = self.menuDispIndic.isChecked()
        self._specInd.mesh.visible = viz
        self._hypInd.mesh.visible = viz
        self._TimeAxis.mesh.visible = viz
        self._fcn_slider_move()

    def _disptog_zoom(self):
        """Toggle zoom mode."""
        active_indic = self.menuDispZoom.isChecked()
        if not active_indic:
            # Spectrogram camera :
            self._speccam.rect = (self._time.min(), self._spec.freq[0],
                                  self._time.max() - self._time.min(),
                                  self._spec.freq[-1] - self._spec.freq[0])
            self._specInd.mesh.visible = self.menuDispIndic.isChecked()
            # Hypnogram camera :
            self._hypcam.rect = (self._time.min(), -5.,
                                 self._time.max() - self._time.min(), 7.)
            # Time camera :
            self._timecam.rect = (self._time.min(), 0.,
                                  self._time.max() - self._time.min(), 1.)

        # Activate / Deactivate indicators :
        self.menuDispIndic.setChecked(not active_indic)
        self.menuDispIndic.setEnabled(not active_indic)
        self._hypInd.mesh.visible = not active_indic
        self._specInd.mesh.visible = not active_indic
        self._TimeAxis.mesh.visible = not active_indic

        self._fcn_slider_settings()



"""Goup of functions for index / event managment."""

import numpy as np

__all__ = ('_events_distance_fill', '_events_to_index', '_index_to_events')


def _events_distance_fill(index, min_distance_ms, sf):
    """Remove events that do not have the good duration.

    Parameters
    ----------
    index : array_like
        Indices of supra-threshold events.
    min_distance_ms : int
        Minimum distance (ms) between two events to consider them as two
        distinct events
    sf : float
        Sampling frequency of the data (Hz)

    Returns
    -------
    f_index : array_like
        Filled (corrected) Indices of supra-threshold events
    """
    # Convert min_distance_ms
    min_distance = min_distance_ms / 1000. * sf
    idx_diff = np.diff(index)
    condition = idx_diff > 1
    idx_distance = np.where(condition)[0]
    distance = idx_diff[condition]
    bad = idx_distance[np.where(distance < min_distance)[0]]
    # Fill gap between events separated with less than min_distance_ms
    if len(bad) > 0:
        fill = np.hstack([np.arange(index[j] + 1, index[j + 1])
                          for i, j in enumerate(bad)])
        f_index = np.sort(np.append(index, fill))
        return f_index
    else:
        return index


def _events_to_index(x):
    """Convert a continuous vector of indices into an 2D array (start, end).

    Parameters
    ----------
    x : array_like
        Array of indices.

    Returns
    -------
    xidx : array_like
        An array of shape (n_events, 2) where the dimension 2 refer to the
        indices where each event start and finish.
    """
    # Split indices where it stopped :
    sp = np.split(x, np.where(np.diff(x) != 1)[0] + 1)
    # Return (start, end) :
    return np.array([[k[0], k[-1]] for k in sp]).astype(int)


def _index_to_events(x):
    """Convert a 2D (start, end) array into a continuous one.

    Parameters
    ----------
    x : array_like
        2D array of indicies.

    Returns
    -------
    index : array_like
        Continuous array of indicies.
    """
    index = np.array([])
    for k in range(x.shape[0]):
        index = np.append(index, np.arange(x[k, 0], x[k, 1] + 1))
    return index.astype(int)
